/*********************************************************************************
 * global_headers.h ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#pragma once
#define _CRT_SECURE_NO_DEPRECATE 1
#define _CRT_NONSTDC_NO_DEPRECATE 1

#if __DTMFBOX_C
#define __EXTERN
#else
#define __EXTERN extern
#endif

// --------------------------------------------------------------------------------------------
// CFLAGS
// --------------------------------------------------------------------------------------------
#include "config_compile.h"

// --------------------------------------------------------------------------------------------
// default defines (when not defined by config_compile.h)
// --------------------------------------------------------------------------------------------
#ifndef DTMFBOX_USE_CAPI
  #define DTMFBOX_USE_CAPI 1
#endif
#ifndef DTMFBOX_USE_VOIP
  #define DTMFBOX_USE_VOIP 1
#endif
#ifndef DTMFBOX_USE_ICE
    #if DTMFBOX_USE_VOIP
        #define DTMFBOX_USE_ICE 1
    #else
        #define DTMFBOX_USE_ICE 0
    #endif
#endif
#ifndef DTMFBOX_USE_SOUND
  #define DTMFBOX_USE_SOUND 1
#endif
#ifndef DTMFBOX_HELP_USAGE
  #define DTMFBOX_HELP_USAGE 1
#endif
#ifndef DTMFBOX_SPEEX_ENCODER
  #define DTMFBOX_SPEEX_ENCODER 0
#endif
#ifndef PJSIP_VERSION
    #define PJSIP_VERSION 8
#endif

#if DTMFBOX_WIN32
    #define DTMFBOX_SCRIPT_IN "dtmfbox_in.tmp"      // Filename for shared memory
    #define DTMFBOX_PIDFILE "dtmfbox.pid"           // Process-ID file
    #define DTMFBOX_CFGFILE "dtmfbox.cfg"           // Config file
    #define DTMFBOX_PIPES "\\\\.\\pipe"             // Pipe (do not change)
#else        
    #define DTMFBOX_SCRIPT_IN   138                 // Magic number for shared memory
    #define DTMFBOX_PIDFILE "/var/run/dtmfbox.pid"  // Process-ID file
    #define DTMFBOX_CFGFILE "./dtmfbox.cfg"         // Config file
    #define DTMFBOX_PIPES "/var/tmp"                // Pipe directory
#endif

// --------------------------------------------------------------------------------------------
// internal defines
// --------------------------------------------------------------------------------------------
#define APPNAME             "dtmfbox"
#define APPVER              "0.5.0-beta1"

#define DEVNULL             "/dev/null"

#define MAX_ACCOUNTS        10                                  // Max. count of VoIP/CAPI accounts
#define MAX_TRANSPORTS      10                                  // Max. count of RTP/RTCP transports
#define MAX_CLIENTS         10                                  // Max. count of VoIP clients, which are able to register (registrar mode)
#define MAX_CONTACTS        10                                  // Max. contact count to save in array
#define INCONF_PORTS        10                                  // Max. count of in_conf ports (in-conference) for a connection

#define MAX_CONS            MAX_CLIENTS + MAX_ACCOUNTS          // Max. count of connection slots (MAX_ACCOUNTS + MAX_CLIENTS)
#define N_CON_PORTS         INCONF_PORTS + 3                    // Mixer array size (10 filestreams + 1 filestream recorder + RX/TX)

#define SOUND_CLOCKRATE     8000                                // Default clockrate for soundcard

#define CAPI_CLOCKRATE      8000                                // Default clockrate for capi
#define CAPI_PACKETS        160                                 // 160 samples (CAPI20_REGISTER -> DATA_B3)
#define CAPI_BUFFERCNT      4                                   // CAPI DATA_B3 buffer blocks
#define CAPI_BUFFERSIZE     CAPI_BUFFERCNT * (CAPI_PACKETS * 2) // CAPI buffer size for samples
#define CAPI_CONTROLLERS    5                                   // Max. number of controllers for LISTEN_REQ 

#define CONF_CLOCKRATE      8000                                // Default clockrate for conference bridge 
#define CONF_CHANNELS       1                                   // 1 channel (mono)
#define CONF_BITS           16                                  // 16 bits
#define CONF_PORTS          256                                 // Max. count of conference ports - 20 connections (media slots) + 20 tone slots + (20 * 10 play slots) = 240. Any more? Hope that fits :)

#define MAX_URI_LENGTH      PJSIP_MAX_URL_SIZE                  // Max. length of SIP-URI or numbers 

#define LOG_BUFFERSIZE      4096                                // Max. size of log buffer 

#define POOL_SIZE_GLOBAL    4096                                // Size for global memory pool
#define POOL_INCR_GLOBAL    1024                                // Increment value for global pool
#define POOL_SIZE_CON       1024                                // Size for connection memory pool
#define POOL_INCR_CON       256                                 // Increment value for connection pool

#define MIX_DEF_RX_VOLUME   100
#define MIX_DEF_TX_VOLUME   100

// --------------------------------------------------------------------------------------------
// default headers
// --------------------------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <fcntl.h>
#include <signal.h>

#if DTMFBOX_WIN32

    #include <windows.h>
    #include <process.h>
    #define usleep(x)               Sleep(x);
    HANDLE shm_id;
    HANDLE shm_map;  

#else
    
    #include <sys/shm.h>
    #include <sys/types.h>
    #include <sys/ipc.h>  
    #include <sys/wait.h>
    #include <sys/time.h>
    #include <sys/resource.h>
    #include <sys/stat.h>
    #include <unistd.h>
    #include <pthread.h>
    #include <sched.h>
    #include <dlfcn.h>

    #define HMODULE void*
    #define LoadLibraryA(a)         dlopen(a, RTLD_LAZY);
    #define GetProcAddress(a, b)    dlsym(a, b);
    #define FreeLibrary(a)          dlclose(a);

    int shm_id;

#endif

// --------------------------------------------------------------------------------------------
// capi headers
// --------------------------------------------------------------------------------------------
#if DTMFBOX_USE_CAPI
    #if DTMFBOX_WIN32

        // Windows
        #define TARGET_32BIT 1
        #include <cwin32.h>

        typedef struct capi_profile
        {
          WORD ncontroller;
          WORD nbchannel;
          DWORD goptions;
          DWORD support1;
          DWORD support2;
          DWORD support3;
          BYTE reserved[24];
          BYTE manu[20];
        } capi_profile;
        
        __declspec(dllimport) unsigned WINAPI CAPI_WAIT_FOR_SIGNAL (unsigned ApplId);
        __declspec(dllimport) unsigned WINAPI CAPI_INSTALLED(void);

        #define CAPI20_ISINSTALLED() CAPI_INSTALLED()
        #define ALERT_REQ(a,b,c,d,e,f,g,h,i) ALERT_REQ(a,b,c,d,e,f,g,h)
        #define INFO_REQ(a,b,c,d,e,f,g,h,i,j) INFO_REQ(a,b,c,d,e,f,g,h,i)
        #define CONNECT_REQ(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x) CONNECT_REQ(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v) 
        #define CONNECT_RESP(a,b,c,d,e,f,g,h,i,k,l,m,n,o,p,q,r,s,t) CONNECT_RESP(a,b,c,d,e,f,g,h,i,k,l,m,n,o,p,q,r,s)
    
    #else

        // Unix
        #include <capi20.h>
        #include <linux/capi.h>

    #endif

#endif

// --------------------------------------------------------------------------------------------
// pjsip headers
// --------------------------------------------------------------------------------------------
#include <pjsip.h>
#include <pjmedia.h>
#include <pjsip_ua.h>
#include <pjlib.h>

#if defined(PJMEDIA_HAS_SPEEX_CODEC) && PJMEDIA_HAS_SPEEX_CODEC!=0 && DTMFBOX_SPEEX_ENCODER!=0
    #include <speex/speex.h>    
#endif

#if DTMFBOX_USE_VOIP
    #include <pjlib-util.h>
    #include <pjmedia-codec.h>
    #include <pjsip_simple.h>
    #include <pjsip-simple/evsub.h>
    #include <pjsip/sip_msg.h>
    #include <pjsip/sip_auth.h>
    #include <pjsip/sip_auth_parser.h>
    #include <pjsip/sip_auth_msg.h>
    #include <pj/except.h>
    #include <pj/rand.h>
    #include <pjlib-util/md5.h>    
    #if DTMFBOX_USE_ICE        
        #include <pjnath/ice_strans.h>
    #endif
#endif

// --------------------------------------------------------------------------------------------
// PJSIP version differences
// --------------------------------------------------------------------------------------------
#if(PJSIP_VERSION == 9)
    #define pj_gethostip(a) pj_gethostip(pj_AF_INET(), a)
    #define pjmedia_transport_udp_close(a) pjmedia_transport_close(a)
#endif

// --------------------------------------------------------------------------------------------
// application typedef structs
// --------------------------------------------------------------------------------------------
#include "type_global.h"                                        // Global type          (oGLB)
#include "type_connection.h"                                    // Connection type      (oCONS[x])
#include "type_settings.h"                                      // Settings type        (oSET)
#include "type_sh_actions.h"                                    // Shared memory type   (sh_actions)

// --------------------------------------------------------------------------------------------
// application includes
// --------------------------------------------------------------------------------------------
#include "dtmfbox.h"                                            // main()
#include "config.h"                                             // Configuration (dtmfbox.cfg)
#include "funcs.h"                                              // Useful functions
#include "options.h"                                            // Parameter parsing options
#include "script_funcs.h"                                       // Script functions
#include "plugins.h"
#include "pipe.h"                                               // Pipe functions
#include "voip_ctrl.h"                                          // VoIP controls (Init, Deinit, Loop, Streaming)
#include "voip_funcs.h"                                         // VoIP functions
#include "voip_auth.h"                                          // VoIP authorization (registrar/client)
#include "voip_registrar.h"                                     // VoIP registrar functions
#include "voip_events.h"                                        // VoIP events 
#include "capi_events.h"                                        // CAPI events
#include "capi_ctrl.h"                                          // CAPI controls (Init, Deinit, Loop, Streaming)
#include "capi_funcs.h"                                         // CAPI functions
#include "alaw_ulaw.h"                                          // a-law/u-law/pcm converting
#include "mixer.h"                                              // Mixer (sound effect managing) - (global getframe/putframe)
#include "mixer_conf.h"                                         // Mixer (conference port managing)
#include "port_capi.h"                                          // CAPI media port (for CAPI and SCRD)
#include "port_voip.h"                                          // VoIP media port 
#include "port_soundcard.h"                                     // SCRD media port (contains a port_capi as subport!)
#include "port_rtpstream.h"                                     // RTP/RTCP streaming port
#include "port_filestream.h"                                    // Filestream media port (wave/raw stream). 
#include "port_tone.h"                                          // Tone media port (fake ringing and custom tones)
#include "port_null.h"                                          // Null port (just to keep the media flow)

#if DTMFBOX_SPEEX_ENCODER != 0
#include "speex.h"                                              // Speex file encoder/decoder
#endif

