/*********************************************************************************
 * mixer_conf.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "mixer_conf.c"

// Connect two connections
pj_status_t mixer_conf_port_connect(connection_t *mCONsrc, connection_t *mCONdst, int add_conf_array)
{
    connection_t *oCONsrc = (connection_t*)mCONsrc;
    connection_t *oCONdst = (connection_t*)mCONdst;
    pj_status_t status = PJ_SUCCESS;

    if(oCONsrc == NULL || oCONdst == NULL)
        return PJMEDIA_ERROR;

    if(oCONsrc->slots->media == -1 || oCONdst->slots->media == -1)
        return PJMEDIA_ERROR;

    // in conference bridge?
    status = mixer_conf_ports_are_connected(oCONsrc, oCONdst);
    if(status != PJ_SUCCESS)
    {
        // connect src with dest (conference bridge)
        status = pjmedia_conf_connect_port(oGLB.conf.bridge, oCONsrc->slots->media, oCONdst->slots->media, 0);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to connect media slots (%d -> %d)", status,  oCONsrc->id, oCONdst->id);
            return status;
        } else {
            //app_dbg3(THIS_FILE, "Connect port %d > %d", oCONsrc->id, oCONdst->id);
        }
    }

    // remember port direction (in_conf)
    if((add_conf_array))    
        mixer_conf_array_add(oCONsrc, oCONdst->id);         
    
    /*
    mixer_conf_dynamic_null_device(oCONsrc);
    mixer_conf_dynamic_null_device(oCONdst);
    */
    return status;
}

// Disconnect two connections
pj_status_t mixer_conf_port_disconnect(connection_t *mCONsrc, connection_t *mCONdst, int remove_conf_array)
{
    connection_t *oCONsrc = (connection_t*)mCONsrc;
    connection_t *oCONdst = (connection_t*)mCONdst;
    pj_status_t status = PJ_SUCCESS;

    if(oCONsrc == NULL || oCONsrc == NULL)
        return PJMEDIA_ERROR;

    if(oCONsrc->slots->media == -1 || oCONdst->slots->media == -1)
        return PJMEDIA_ERROR;

    // in conference bridge?
    status = mixer_conf_ports_are_connected(oCONsrc, oCONdst);
    if(status == PJ_SUCCESS)
    {
        // disconnect src from dest (conference bridge)
        status = pjmedia_conf_disconnect_port(oGLB.conf.bridge, oCONsrc->slots->media, oCONdst->slots->media);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to disconnect media slots (%d -> %d)", status,  oCONsrc->id, oCONdst->id);
            return status;
        } 
        else 
        { 
            //app_dbg(THIS_FILE, "Disconnect port %d > %d", oCONsrc->id, oCONdst->id);
        }      
    }

    // remove from port direction (in_conf)
    if((remove_conf_array))
        mixer_conf_array_remove(oCONsrc, oCONdst->id);    

    /*
    mixer_conf_dynamic_null_device(oCONsrc);
    mixer_conf_dynamic_null_device(oCONdst);
    */

    return status;
}

// Disconnect the slot from all other media_slots
pj_status_t mixer_conf_port_disconnect_all(int slot)
{
    pjmedia_conf_port_info info[MAX_CONS];
    unsigned int i;
    unsigned int count;
    unsigned j;
    pj_status_t status = PJ_SUCCESS;
    int con_id = -1;

    if(slot == -1)
        return PJ_TRUE;

    // find con-id to slot
    for(i=0; i<(unsigned int)oSET.max_connections; i++)
    {
        if(oCONS[i].slots->media == slot) {
            con_id = i;
            break;
        }
    }

    // disconnect connection from in_conf array (and conference bridge)
    if(con_id != -1)
    {   
        for(i=0; i<(unsigned int)oSET.max_connections; i++)
        {
            for(j=0; j<INCONF_PORTS; j++)
            {
                if(oCONS[i].in_conf[j] == con_id) {
                    mixer_conf_port_disconnect(&oCONS[i], &oCONS[con_id], 1);
                    mixer_conf_port_disconnect(&oCONS[con_id], &oCONS[i], 1);
                }
            }
        }         
    }

    // disconnect slot from each other slot (maybe other port type, e.g. tone-slot?)
    count = PJ_ARRAY_SIZE(info);
    pjmedia_conf_get_ports_info(oGLB.conf.bridge, &count, info);
    for (i=0; i<count; i++)
    {
        pjmedia_conf_port_info *port_info = &info[i];
        for(j=0; j<port_info->listener_cnt; j++)
        {
            if ((int)port_info->listener_slots[j] == slot)
            {
                status = pjmedia_conf_disconnect_port(oGLB.conf.bridge, slot, port_info->slot);
                status = pjmedia_conf_disconnect_port(oGLB.conf.bridge, port_info->slot, slot);                    
            }
                
            if((int)port_info->slot == slot)
            {
                status = pjmedia_conf_disconnect_port(oGLB.conf.bridge, slot, port_info->listener_slots[j]);
                status = pjmedia_conf_disconnect_port(oGLB.conf.bridge, port_info->listener_slots[j], slot);                    
            }                             
        }
    }
    
    return status;
}

/*
static pj_status_t mixer_conf_dynamic_null_device(connection_t *mCON)
{
    pjmedia_conf_port_info info[CONF_PORTS];    
    connection_t *CUR_CON = (connection_t*)mCON;
    unsigned int own_media_slot = CUR_CON->slots->media;
    unsigned int own_listeners = 0;
    unsigned int count = CONF_PORTS;
    int got_getframe = 0;
    int got_putframe = 0;
    unsigned int x, y;

    if(own_media_slot != -1 && oGLB.conf.bridge != NULL)
    {
        //pjmedia_conf_disconnect_port(oGLB.conf.bridge, CUR_CON->slots->media, oGLB.audio.flow_slot);
        //pjmedia_conf_disconnect_port(oGLB.conf.bridge, oGLB.audio.flow_slot, CUR_CON->slots->media);

        // check, if we got any listeners
        if(pjmedia_conf_get_port_info(oGLB.conf.bridge, own_media_slot, &info[0]) == PJ_SUCCESS)
        {            
            own_listeners = info[0].listener_cnt;            
            for(x=0; x<info[0].listener_cnt; x++)
            {
                if(info[0].listener_slots[x] != oGLB.audio.flow_slot)
                    got_putframe = 1;
            }

            //if(own_listeners > 0) {
              //  got_putframe = 1;
            //}
        }

        // check, if we are a listener
        count = PJ_ARRAY_SIZE(info);
        if(pjmedia_conf_get_ports_info(oGLB.conf.bridge, &count, info) == PJ_SUCCESS) 
        {
            for(x=0; x<count; x++)
            {
                pjmedia_conf_port_info *port_info = &info[x];
                
                if(port_info->slot != oGLB.audio.flow_slot)
                {
                    for(y=0; y<port_info->listener_cnt; y++)
                    {
                        if(port_info->listener_slots[y] == own_media_slot)
                        {
                            got_getframe = 1;
                            break;
                        }                   
                    }

                    if(got_getframe) 
                        break;
                }
            }
        }
        
        if(!got_putframe)            
            pjmedia_conf_connect_port(oGLB.conf.bridge, CUR_CON->slots->media, oGLB.audio.flow_slot, 0);
        else
            pjmedia_conf_disconnect_port(oGLB.conf.bridge, CUR_CON->slots->media, oGLB.audio.flow_slot);

        if(!got_getframe)                       
            pjmedia_conf_connect_port(oGLB.conf.bridge, oGLB.audio.flow_slot, CUR_CON->slots->media, 0);                
        else          
            pjmedia_conf_disconnect_port(oGLB.conf.bridge, oGLB.audio.flow_slot, CUR_CON->slots->media);
    }

    return PJ_SUCCESS;
}
*/

static pj_status_t mixer_conf_slots_are_connected(unsigned int src, unsigned int dst)
{
    pjmedia_conf_port_info info;    
    pj_status_t status = PJMEDIA_ERROR;
    unsigned int j;
        
    if(pjmedia_conf_get_port_info(oGLB.conf.bridge, src, &info) == PJ_SUCCESS)
    {
        for(j=0; j<info.listener_cnt; j++)
        {
            if ((int)info.listener_slots[j] == dst)
            {
                // found!
                status = PJ_SUCCESS;
                break;
            }
        }
    }

    return status;
}

// check if two media ports are connected over conference bridge.
// return = PJ_SUCCESS, ports are connected
static pj_status_t mixer_conf_ports_are_connected(connection_t *mCONsrc, connection_t *mCONdst)
{    
//    connection_t *oCONsrc = (connection_t*)mCONsrc;
//    connection_t *oCONdst = (connection_t*)mCONdst;            
            
    if(mCONsrc->slots->media == -1 || mCONdst->slots->media == -1 || oGLB.conf.bridge == NULL)
        return PJMEDIA_ERROR;

    return mixer_conf_slots_are_connected(mCONsrc->slots->media, mCONdst->slots->media);    
}

// add connection to conference array (oCONS[x].in_conf)
pj_status_t mixer_conf_array_add(connection_t *mCON, int con_id)
{
    connection_t *CUR_CON = (connection_t*)mCON;
    int i;

    if(CUR_CON->in_conf_cnt+1 > INCONF_PORTS)
    {
        app_err(THIS_FILE, "No more conference ports left (max: %d)!", INCONF_PORTS);
        return 1;
    }

    // check, if con_id exists..
    if(con_id >= 0 && con_id < oSET.max_connections)
        if(oCONS[con_id].info->type == FREE)
            return 1;

    // check, that not already exists!
    for(i=0; i<INCONF_PORTS; i++)
    {
        if(CUR_CON->in_conf[i] == con_id)
            return 1;
    }

    // add to in_conf array
    for(i=0; i<INCONF_PORTS; i++)
    {
        if(CUR_CON->in_conf[i] == -1)
        {
            CUR_CON->in_conf[i] = con_id;
            CUR_CON->in_conf_cnt++;
            break;
        }
    }

    return PJ_SUCCESS;
}

// remove connection from conference array (oCONS[x].in_conf)
pj_status_t mixer_conf_array_remove(connection_t *mCON, int con_id)
{
    connection_t *CUR_CON = (connection_t*)mCON;
    int tmp[INCONF_PORTS];
    int i;
    
    // remove connection from in_conf
    for(i=0; i<INCONF_PORTS; i++)
    {
        if(CUR_CON->in_conf[i] == con_id)
        {
            CUR_CON->in_conf[i] = -1;            
        }
    }

    // recount in_conf_cnt and reorder
    CUR_CON->in_conf_cnt = 0;
    for(i=0; i<INCONF_PORTS; i++)
    {
        if(CUR_CON->in_conf[i] != -1)
        {
            tmp[CUR_CON->in_conf_cnt] = CUR_CON->in_conf[i];
            CUR_CON->in_conf_cnt++;
        }
    }

    // set new order
    for(i=0; i<CUR_CON->in_conf_cnt; i++)
        CUR_CON->in_conf[i] = tmp[i];

    for(i=CUR_CON->in_conf_cnt; i<INCONF_PORTS; i++)
        CUR_CON->in_conf[i] = -1;

    return PJ_SUCCESS;
}
