#undef __DTMFBOX_MENU_PLUGIN_C
#include "menu.plugin.h"
#include "menu.h"
#include "config.h"

static char* get_value(char *file, char* section, char* key, char* value);
//static char *get_sep(char *val_in, char *val_out, int size, char split_char);
static char* rtrim(char *str);
static char* ltrim(char *str);

// Read INI values by section and key
static char* get_value(char *file, char* section, char* key, char* value)
{
    char line[1024];
    char* findsec = strdup(section);
    char* findkey = strdup(key);
    FILE* fp;
    char* pos;
    int insection;
    int found;
    unsigned int i;
    
    // Open file
    fp = fopen(file, "r");

    // error - cannot read file!
    if(fp == NULL) {
        free(findkey); free(findsec);
        findkey = NULL; findsec = NULL;
        return NULL;
    }

    //convert section & key to uppercase
    for(i=0; i<strlen(findsec); i++)
        findsec[i] = (char)toupper(findsec[i]);
    for(i=0; i<strlen(findkey); i++)
        findkey[i] = (char)toupper(findkey[i]);

    insection = 0;
    found = 0;

    // Each line
    while(fgets(line, sizeof(line), fp) != NULL)
    {
        // remove comments
        pos = strstr(line, "//");
        while(pos != NULL) {
            int no_comment = 0;
            if(pos - line != 0)
            {
                if(line[(pos - line)-1] == 92) {
                    no_comment = 1;
                    memcpy(line + ((pos - line)-1), line + ((pos - line)), strlen(pos));                    
                }
            }

            if(!no_comment)
                pos[0] = '\0';

            pos = strstr(pos+2, "//");
        }

        // remove CR
        pos = strchr(line, '\n');
        if(pos != NULL)
            pos[0] = '\0';

        // trim
        ltrim(line); rtrim(line);

        // copy return value
        if(found==0)
            strcpy(value, line);

        // convert line to uppercase
        for(i=0; i<strlen(line); i++)
            line[i] = (char)toupper(line[i]);

        if(insection == 0)
        {
            // enter section
            pos = strstr(line, findsec);
            if(pos!=NULL)
            {
                insection = 1;
            }
        } else {
            // leave section
            if(line[0] == '[')
            {
                insection = 0;
            }
        }

        if(insection == 1)
        {
            pos = strstr(line, findkey);
            if(pos!=NULL)
            {
                // key found!!
                found = 1;
                strcpy(value, value + strlen(findkey));
                break;
            }
        }
    }

    // Close file
    fclose(fp);

    // nothing found?
    if(found == 0)
        value[0] = '\0';

    // free
    free(findkey); free(findsec);
    findkey = NULL; findsec = NULL;

    return NULL;
}


char **cfg_get_entries(char *file, int type, char* section, unsigned *entry_count, char **data)
{
    char line[1024];    
    char* findsec = strdup(section);
    FILE* fp;
    char* pos;
    int insection;
    int found;
    int entry_pos = 0;
    
    // Open file
    fp = fopen(file, "r");

    if(data != NULL) {
        free(findsec); findsec = NULL;
        return data;
    }

    // error - cannot read file!
    if(fp == NULL) {
        free(findsec); findsec = NULL;
        return data;
    }

    insection = 0;
    found = 0;

    // Each line
    while(fgets(line, sizeof(line), fp) != NULL)
    {
        // remove comments
        pos = strstr(line, "//");
        while(pos != NULL) {
            int no_comment = 0;
            if(pos - line != 0)
            {
                if(line[(pos - line)-1] == 92) {
                    no_comment = 1;
                    memcpy(line + ((pos - line)-1), line + ((pos - line)), strlen(pos));                    
                }
            }

            if(!no_comment)
                pos[0] = '\0';

            pos = strstr(pos+2, "//");
        }

        // remove CR
        pos = strchr(line, '\n');
        if(pos != NULL)
            pos[0] = '\0';

        // trim
        ltrim(line); rtrim(line);

        if(insection == 0)
        {
            // enter section
            pos = strstr(line, findsec);
            if(pos!=NULL)
            {
                insection = 1;
            }
        } else {
            // leave section
            if(line[0] == '[')
            {
                insection = 0;
                break;
            }
        }

        if(insection == 1)
        {
            pos = strstr(line, "=");
            
            if(pos!=NULL)
            {
                char *pos_val;
                int len = (int)(pos-line);
                if(len < 0) len = 0;

                // key found!!
                found = 1;
                if(type == 0) 
                {
                    // get key (KEY=...)
                    data = (char**)realloc(data, sizeof(char*) * (entry_pos+1));
                    data[entry_pos] = malloc(len+1);                                
                    memcpy(data[entry_pos], line, len);
                    data[entry_pos][len] = '\0';
                } 
                else 
                {    
                    // get value (...=VALUE)
                    pos_val = strstr(line, "=");
                    if(pos_val!=NULL)
                    {                     
                        int val_len = (int)strlen(pos_val)-1; 
                        if(val_len<0) val_len = 0;

                        data = (char**)realloc(data, sizeof(char*) * (entry_pos+1));
                        data[entry_pos] = malloc(val_len+1);                    
                        memcpy(data[entry_pos], pos_val+1, val_len);
                        data[entry_pos][val_len] = '\0';                    
                    }
                }

                entry_pos++;
            }
        }
    }

    // Close file
    fclose(fp);

    // free
    free(findsec); findsec = NULL;

    *entry_count = entry_pos;
    return data;
}

// Return all sections from a file
char **cfg_get_sections(char *file, int *sec_count, char **sections)
{
    char line[1024];        
    FILE* fp;
    char* pos;        
    int sec_pos = 0;
    char* sec_start;    
    char* sec_end;
    
    // make sure, sections is null. It gets allocated here.
    if(sections != NULL) return sections;
        
    // Open file
    fp = fopen(file, "r");    
    if(fp == NULL) return sections;
    
    // Each line
    while(fgets(line, sizeof(line), fp) != NULL)
    {
        // remove comments
        pos = strstr(line, "//");
        while(pos != NULL) {
            int no_comment = 0;
            if(pos - line != 0)
            {
                if(line[(pos - line)-1] == 92) {
                    no_comment = 1;
                    memcpy(line + ((pos - line)-1), line + ((pos - line)), strlen(pos));                    
                }
            }

            if(!no_comment)
                pos[0] = '\0';

            pos = strstr(pos+2, "//");
        }

        // remove CR
        pos = strchr(line, '\n');
        if(pos != NULL)
            pos[0] = '\0';

        // trim
        ltrim(line); rtrim(line);

        sec_start = strchr(line, '[');
        sec_end = strchr(line, ']');

        if(sec_start != NULL && sec_end != NULL)
        {
            if(sec_start < sec_end)
            {
                int len = (int)(sec_end - sec_start);

                if(len > 0) {
                    sections = (char**)realloc(sections, sizeof(char*) * (sec_pos + 1));
                    sections[sec_pos] = malloc(len);
                    memcpy(sections[sec_pos], line+1, len-1);
                    sections[sec_pos][len-1] = '\0';
                    sec_pos++;                    
                }
            }
        }
    }
        
    // Close file
    fclose(fp);

    *sec_count = sec_pos;
    return sections;
}

// split values 
// val_in       = input value
// val_out      = output value (also return)
// size        = sizeof(val_out)
// split_char   = splitting char
/*
static char *get_sep(char *val_in, char *val_out, int size, char split_char)
{
    char *p;    
        
    memset(val_out, 0, size);

    p = strchr(val_in, split_char);
    if(p != NULL)
    {
        int psize = (int)(p - val_in);
        if(psize >  size)
            psize = sizeof(val_out);

        strncpy(val_out, val_in, psize);
        val_out[psize] = '\0';

        strcpy(val_in, p+1);        
    }
    else
    {        
        if(strlen(val_in))
        {
            strncpy(val_out, val_in, strlen(val_in));    
            val_out[strlen(val_in)] = '\0';

            strcpy(val_in, "");
        } 
        else
        {
            strcpy(val_out, "");
        }
    }    

    ltrim(val_out);
    rtrim(val_out);

    return val_out;
}
*/

int cfg_strlen(char *str)
{
    int i = 0;

    while(1)
    {
        if(((unsigned)str[i] > 9 && (unsigned)str[i] <= 127))        
            i++;
        else
            break;        
    }

    return i;
}


// Right-Trim
static char* rtrim(char *str)
{
    char *s, *p;
    unsigned char c;
    s = p = str;
    while ((c = *s++))
        if (!isspace(c)) p = s;
            *p = '\0';
    return str;
}

// Left-Trim
static char* ltrim(char *str)
{
    char *s, *p;
    for (s = p = str; *s && isspace(*s); s++);
        while ((*p++ = *s++));

    return str;
}

// Check, that string is numerical
int cfg_isnumeric(char *str, int check_decimal)
{
    unsigned int i;
    int ret = -1;
    
    if(str == NULL)
        return 0;

    ret = (int)strlen(str);

    for(i=0; i<strlen(str); i++)
    {
        if((int)str[i] >= 0 && (int)str[i] <= 255)
        {
            int digit = isdigit(str[i]);
            
            if(!digit && check_decimal)
                if(str[i] == '.')
                    digit = 1;

            if(!digit)
            {
                ret = 0;
                break;
            }
        }
    }

    return ret;
}


// cfg_replace a string
char* cfg_replace(const char *in, char *out, int size, const char* find, const char* cfg_replace)
{
	const char* p;
    char *ret;
    char *tmp;
    int pos = 0;

    ret = malloc(size);
    tmp = malloc(size);

    if (out == NULL || ret == NULL || tmp == NULL) {
        if(ret) free(ret);
        if(tmp) free(tmp);
        tmp = NULL; ret = NULL;
		return NULL;
    } else {
        memset(ret, 0, size);
        memset(tmp, 0, size);
        strncpy(tmp, in, size);
    }
	
    p = tmp;
    do
    {
        if(p!=NULL) p = strstr(p, find);
        if(p == NULL) break;
        pos = (int)(strlen(in) - strlen(p));

        // copy first part
        if (strcmp(p, in)) {
		    strncpy(ret, tmp, p-tmp);
        }

        // cfg_replace with cfg_replace-string and strcat the rest
	    strcat(ret + (p-tmp), cfg_replace);
	    strcat(ret, p + strlen(find));

        // save as new text
        strncpy(tmp, ret, size);
        
        // research from beginning
        p = tmp;
        memset(ret, 0, size);
        
    } while(p != NULL);

    strncpy(out, tmp, size);
    free(ret); free(tmp);
    tmp = NULL; ret = NULL;

	return out;
}

void read_plugin_config()
{
    char value[1024];
    char section[64];    
    int i;
    
    free_plugin_config();

    // [audio]
    strcpy(section, "[audio]");
    get_value(PLUGIN_CFG, section, "text2speech=", value);
    memcpy(sets.text2speech.cmd, value, sizeof(sets.text2speech.cmd));    

    // [accX]
    for(i=0; i<MAX_ACC; i++)
    {        
        sprintf(section, "[acc%d]", i+1);
        get_value(PLUGIN_CFG, section, "menu_keycode=", value);
        memcpy(sets.accounts[i].menu_keycode, value, sizeof(sets.accounts[i].menu_keycode));

        get_value(PLUGIN_CFG, section, "menu_entry=", value);
        memcpy(sets.accounts[i].menu_entry, value, sizeof(sets.accounts[i].menu_entry));                
    }
}

void free_plugin_config()
{
    memset(&sets.accounts, 0, sizeof(sets.accounts));
    memset(&sets.text2speech, 0, sizeof(sets.text2speech));
}
