#define PLUGIN_CFG          "dtmfbox.cfg"
#define PLUGIN_MENU_CFG     "menu.cfg"

#define MAX_ACC     10      // max. accounts
#define MAX_CONS    20      // max. connections
#define MAX_ARGS    32      // max. arguments

/* ------------------------------------------------------------------------------------------------------------ */

/*
    oMENU[y]->menu
    This struct contains all keys, values and parameters of a section (menu.cfg).
*/
typedef struct key_tree_t
{   
    char         **keys;
    char         **values;
    int          *link;
    char         section[64];
    unsigned     count;

} key_tree_t;

/* 
    oMENU[y]->lib (lib:)
    For calling library functions 
*/
typedef struct menu_lib_t
{
    char    *library;
    char    *function;
    char    *params[MAX_ARGS];
    int     param_len[MAX_ARGS];    

} menu_lib_t;

/* 
    oMENU[y]->script (script:)
    For calling scripts / batch files / external apps
*/
typedef struct menu_script_t
{    
    char    *cmd;

} menu_script_t;


/* 
    oMENU[y]->actions (action:)
    Contains all actions defined in dtmfbox.cfg
*/
typedef struct menu_action_t
{    
    char                *s_event;
    char                *s_direction;
    char                *s_type;
    char                *s_accid;
    char                *s_nosrc;
    char                *s_nodst;
    char                *s_idsrc;
    char                *s_iddst;
    char                *s_data;
    char                *s_action;

    int                 i_idsrc;
    int                 i_iddst;
    int                 i_accid;
                
    char                *params[MAX_ARGS];
    int                 param_len[MAX_ARGS];    

} menu_action_t;

/* 
    oMENU[y] - Complete menu structure (y=MenuItem)
    The "Start-Menu" for the account is defined in dtmfbox.cfg under the 
    account section (e.g. menu_entry=menu:main).    
*/
typedef struct menu_t
{    
    int                 menu_id;                // The MenuItem-ID (y)

    key_tree_t          *menu;                  // The complete structure from menu.cfg for this menu
    menu_lib_t          *lib;                   // Not null, when the menu is a [lib:...] section
    menu_script_t       *script;                // Not null, when the menu is a [script:...] section
    menu_action_t       *action;                // Not null, when the menu is a [action:...] section

    char                *command;               // The last command
    char                **params;               // Parameters from the last command (max. 32)
    int                 *param_len;             // Parameter length from the last commands (max. 32)
    char                *say;                   // When entering this menu, the message is passed to the text2speech command
    char                *redirect_to;           // Automatically redirect to a menu:, script:, lib:
    int                 redirect_link;          // The MenuItem-ID for redirect

    int                 type;                   // Type: menu:, lib: or script: section

} menu_t;

/* ------------------------------------------------------------------------------------------------------------ */

/* 
    cons[x].status 
    Status struct for a connection. This struct won't be cleared as long as the connection is active.
*/
typedef struct dtmfbox_status
{       
    menu_t              *last_menu;
    menu_t              *current_menu;
    char                dtmf_buffer[256];
    int                 dtmf_pos;
    int                 active;
    int                 is_internal;    
    
    char                tmp_command[256];           // temporary command 
    char                *tmp_params[MAX_ARGS];      // temporary parameters 
    int                 tmp_param_len[MAX_ARGS];

} dtmfbox_status;

/* 
    cons[x] 
    This struct contains the current event infos of a specific connection    
*/
typedef struct dtmfbox_connection
{   
    dtmfbox_status      status;
        
    char                no_src[256];
    char                no_dst[256];    
    char                s_event[32];
    char                data[32];
    char                s_direction[32];
    char                s_type[32];    
    int                 conid_src;
    int                 conid_dst;
    int                 acc_id;    
    
} dtmfbox_connection; 

/* ------------------------------------------------------------------------------------------------------------ */

/* 
    sets.accounts[x] 
    Additional account information for this module (from dtmfbox.cfg)
*/
typedef struct dtmfbox_accounts
{
    char                menu_entry[64];
    char                menu_keycode[16];    
    
} dtmfbox_accounts;

/* 
    sets.text2speech 
    Additional section [text2speech] for this module (from dtmfbox.cfg)
    The text to speech is passed to the command.
*/
typedef struct dtmfbox_text2speech
{
    char                cmd[2048];

} dtmfbox_text2speech;

/* 
    sets  
    Global settings struct for this module
*/
typedef struct dtmfbox_settings
{   
    dtmfbox_accounts    accounts[MAX_ACC];
    dtmfbox_text2speech text2speech;   
    
} dtmfbox_settings;
