#define __DTMFBOX_MENU_PLUGIN_C
#include "menu.plugin.h"
#include "funcs.h"
#include "menu.h"
#include "config.h"

/* Callback gets called when a new event arrives */
EXPORT void plugin_event(char *scriptfile, char *s_event, char *s_direction, char *s_type, int conid_src, int conid_dst, char *no_src, char *no_dst, char *data, int acc_id) 
{   
    char s_log[256];
    char s_id[32];
    int id = conid_src;    
    unsigned i;

    // abort on unknown account. reduce by one to fit array
    acc_id--;
    if(acc_id < 0 || acc_id >= MAX_ACC) {
        if(cons[id].acc_id != -1)
            acc_id = cons[id].acc_id;
        else
            return;
    }
    
    sprintf(s_id, "%d", id);    
    if(id != -1)
    {   
        dtmfbox_status status;
        int next_menu = -1;

        // keep 'status'. So make a backup and restore it after reset!        
        memcpy(&status, &cons[id].status, sizeof(dtmfbox_status)); 
        memset(&cons[id], 0, sizeof(cons[id]));                   
        memcpy(&cons[id].status, &status, sizeof(dtmfbox_status)); 

        // fill connection info
        memcpy(cons[id].s_event, s_event, sizeof(cons[id].s_event));
        memcpy(cons[id].s_type, s_type, sizeof(cons[id].s_type));
        memcpy(cons[id].s_direction, s_direction, sizeof(cons[id].s_direction));
        memcpy(cons[id].data, data, sizeof(cons[id].data));
        memcpy(cons[id].no_src, no_src, sizeof(cons[id].no_src));
        memcpy(cons[id].no_dst, no_dst, sizeof(cons[id].no_dst));                
        cons[id].conid_src = conid_src;
        cons[id].conid_dst = conid_dst;        
        cons[id].acc_id = acc_id;  

        // ------------------------------------------------------------------------------------------------------------------------------------
        // CUSTOM ACTIONS
        // ------------------------------------------------------------------------------------------------------------------------------------        
        if(acc_id != -1)
        {
            for(i=0; i<menu_count; i++)
            {
                if(menus[i]->action)
                {
                    int ok = 0;
                    int not_ok = 0;
                    menu_action_t *action = menus[i]->action;

                    if(action->s_event != NULL) {           
                        if(!strcmp(s_event,action->s_event)) ok++; else not_ok++; 
                    }

                    if(action->s_type != NULL) {
                        if(!strcmp(s_type, action->s_type)) ok++; else not_ok++;
                    }
                   
                    if(action->s_direction != NULL) {
                        if(!strcmp(s_direction, action->s_direction)) ok++; else not_ok++;
                    }
                   
                    if(action->s_nosrc != NULL) {
                        if(!strcmp(no_src, action->s_nosrc)) ok++; else not_ok++;
                    }

                    if(action->s_nodst != NULL) {
                        if(!strcmp(no_dst, action->s_nodst)) ok++; else not_ok++;
                    }

                    if(action->s_accid != NULL) {
                        if(acc_id == action->i_accid) ok++; else not_ok++;
                    }

                    if(action->s_idsrc != NULL) {
                        if(conid_src == action->i_idsrc) ok++; else not_ok++;
                    }

                    if(action->s_iddst != NULL) {
                        if(conid_dst== action->i_iddst) ok++; else not_ok++;
                    }

                    // execute action!
                    if((ok && !not_ok) && action->s_action != NULL)
                    {                           
                        int link_id = search_link(action->s_action, menu_count);

                        if(link_id != -1)
                        {
                            char s_log[256];
                            menu_t *m = menus[link_id];                            

                            m->params = action->params;
                            m->param_len = action->param_len;
                            cons[id].status.current_menu = m;
                                     
                            // Status
                            snprintf(s_log, sizeof(s_log), "[plugin.menu] (#%d): action > %s", cons[id].acc_id+1, action->s_action);
                            app_dbg(s_log);

                            // ...call external!
                            call_external(acc_id, id, cons[id].status.current_menu);
                        }
                    }
                }
            }
        }

        // ------------------------------------------------------------------------------------------------------------------------------------
        // INTERNAL DIAL (DDI) - CONNECTION IS NOT ACTIVE!!
        // ------------------------------------------------------------------------------------------------------------------------------------
        if((!strcmp(s_event, "DDI") || !strcmp(s_event, "CONNECT")) && !strcmp(s_direction, "OUTGOING") && 
            !cons[id].status.active)
        {            
            char *is_sip = strstr((const char*)no_dst, "@");
            char dest_no[256];

            if(is_sip) {
                memset(dest_no, 0, sizeof(dest_no));
                memcpy(dest_no, no_dst, is_sip-no_dst);                
            } else {
                memset(dest_no, 0, sizeof(dest_no));
                memcpy(dest_no, no_dst, sizeof(dest_no));
            }

            if(!strcmp(dest_no, sets.accounts[acc_id].menu_keycode) &&
                strlen(sets.accounts[acc_id].menu_keycode) &&
                strlen(sets.accounts[acc_id].menu_entry)) 
            {   
                int link_id = search_link(sets.accounts[acc_id].menu_entry, menu_count);

                if(link_id != -1) {

                    // Log
                    snprintf(s_log, sizeof(s_log), "[plugin.menu] (#%d): %s (internal)", acc_id+1, menus[link_id]->menu->section);
                    app_msg(s_log);

                    // Hook up
                    dtmfbox(3, s_id, "-hook", "up");

                    // ok!
                    //next_menu = link_id;
                    cons[id].status.active = 1;                             // Active (we take control)
                    cons[id].status.is_internal = 1;                        // Internal, typed keycode (e.g. menu_keycode=600)
                    cons[id].status.last_menu = menus[link_id];
                    cons[id].status.current_menu = menus[link_id];          // set the first menu map (e.g. menu_entry=menu:main)
                    return;
                }
            }
        }

        // ------------------------------------------------------------------------------------------------------------------------------------
        // MENU - CONNECTION IS ACTIVE!
        // ------------------------------------------------------------------------------------------------------------------------------------
        if(cons[id].status.active)
        {
            menu_t *m = cons[id].status.current_menu;
 
            // First entry, after hook up. 
            // Set Entry-Menue
            if(!strcmp(s_event, "CONFIRMED") && m) {
                next_menu = m->menu_id;                
            }

            // Change menu or request a command,
            // because a DTMF digit was entered...
            if(!strcmp(s_event, "DTMF")) 
            {
                unsigned i;
                
                // .. save dtmf character... maybe two or more digits, incomplete or with delimiter..
                int dtmf_pos = cons[id].status.dtmf_pos;
                
                // buffer overflow -> reset!
                if(dtmf_pos + strlen(data) >= sizeof(cons[id].status.dtmf_buffer)) {                    
                    memset(cons[id].status.dtmf_buffer, 0, sizeof(cons[id].status.dtmf_buffer));
                    cons[id].status.dtmf_pos = 0;
                }
                // add dtmf digits to buffer
                if(dtmf_pos + strlen(data) < sizeof(cons[id].status.dtmf_buffer)) {
                    memcpy(&cons[id].status.dtmf_buffer[dtmf_pos], data, strlen(data));                    
                    cons[id].status.dtmf_pos += (int)strlen(data);
                }
                
                // search for menu-link by dtmf.
                // search in current data (single digit) and in dtmf_buffer (multiple digits)
                for(i=0; i<m->menu->count; i++)
                {
                    if(strstr(cons[id].status.dtmf_buffer, m->menu->keys[i]) != NULL || 
                       !strcmp(m->menu->keys[i], data))
                    {                           
                        menu_t *m_old = cons[id].status.current_menu;                        

                        // Get next menu
                        next_menu = m->menu->link[i];
                        
                        // Change current menu to next menu
                        if(next_menu != -1)
                        {   
                            // Set command / parameters from last menu to current menu
                            cons[id].status.last_menu = cons[id].status.current_menu;
                            cons[id].status.current_menu = menus[next_menu];
                            m = cons[id].status.current_menu;                        
                            m->command = m_old->menu->values[i];
                            //m->params = m_old->menu->params[i];
                            //m->param_len = m_old->menu->param_len[i];

                            m->params = cons[id].status.tmp_params;
                            m->param_len = cons[id].status.tmp_param_len;

                            // new parameters in key value? 
                            search_param(m_old->menu->values[i], NULL, 0, m->params, m->param_len, 1);
                        }
                        break;
                    }
                }                
            }

            // -goto command was used. 
            // force selecting the next menue...
            if(!strcmp(s_event, "GOTO")) {
                m = cons[id].status.current_menu;
                next_menu = m->menu_id;
                memset(cons[id].status.tmp_params, 0, sizeof(cons[id].status.tmp_params));
                memset(cons[id].status.tmp_param_len, 0, sizeof(cons[id].status.tmp_param_len));
                snprintf(cons[id].status.tmp_command, sizeof(cons[id].status.tmp_command), "%s", data);

                m->command = cons[id].status.tmp_command;
                m->params = cons[id].status.tmp_params;
                m->param_len = cons[id].status.tmp_param_len;
                
                search_param(m->command, NULL, 0, m->params, m->param_len, 1);
            }
            
            // ...goto a menu or run command??
            if(next_menu != -1) 
            {
                // Log
                snprintf(s_log, sizeof(s_log), "[plugin.menu] (#%d): next > %s", acc_id+1, cons[id].status.current_menu->menu->section);
                app_dbg(s_log);

                // call script or lib
                call_external(acc_id, id, cons[id].status.current_menu);            

                // reset dtmf_buffer
                cons[id].status.dtmf_pos = 0;
                memset(cons[id].status.dtmf_buffer, 0, sizeof(cons[id].status.dtmf_buffer));
            }          
        }
        
        // reset status on disconnect
        if(!strcmp(s_event, "DISCONNECT")) {
            memset(&cons[id].status, 0, sizeof(dtmfbox_status));
        }
    }
   
}

/* Plugin is loaded! */
EXPORT void plugin_init(plugin_exec_cmd cmd) 
{
    int i;

    // get the callback to execute dtmfbox commands
    dtmfbox = cmd;  

    for(i=0; i<MAX_CONS; i++)    
        memset(&cons[i], 0, sizeof(dtmfbox_connection));    

    menu_count = 0;
    
    // read configuration and menu structure
    read_plugin_config();
    read_menu();

    app_msg("[plugin.menu]: Initialized!");
}

/* Plugin is unloaded! */
EXPORT void plugin_close() 
{   
    free_plugin_config();
    free_menu();    
    
    app_msg("[plugin.menu]: Destroyed!");
    dtmfbox = NULL;
}

/* Callback gets called, when dtmfbox outputs something (connection id, -call, -list, ...) */
/*
EXPORT void plugin_output(char *output, int pid) 
{ 
    //printf("%s\n", output);    
}
*/

/* Callback gets called, when command help is shown */
EXPORT char *plugin_extra_help()
{
    return " dtmfbox menu v0.1\n\n" \
           "  -goto \"menu:X|lib:X|script:X\"         - goto menu\n" \
           "  -stop menu                            - leave menu\n" \
           "  -speak \"text\"                         - say text (text2speech)\n";
}

EXPORT int plugin_extra_commands(int argc, char *arg[], int pid)
{
    int i;
    int id = -1;
    int ret = 0;

    // Validate Connection-ID
    if(cfg_isnumeric(arg[1], 0))
    {
        id = atoi(arg[1]);
        if(strcmp(arg[1], "0") && id == 0)
            id = -1;        
    } 

    // Parse parameters
    for(i=1; i<argc; i++)
    {
        if(id != -1)
        {
            // -speak "text"
            if(!strcmp(arg[i], "-speak") && argc > i+1) {      
                speak(id, arg[i+1]);        
                ret++;
            }

            // -stop menu
            if(!strcmp(arg[i], "-stop") && argc > i+1) {
                if(!strcmp(arg[i+1], "menu")) {
                    cons[id].status.active = 0;
                    ret++;
                }
            }

            // -goto [menu: | lib: | script:]
            if(!strcmp(arg[i], "-goto") && argc > i+1)
            {
                int link_id = -1;
                char command[256];       
                char *pos;
           
                snprintf(command, sizeof(command), "%s", arg[i+1]);
                pos = strstr(command, "(");
                if(pos) pos[0] = '\0';

                // search for link and change menue                        
                link_id = search_link(command, menu_count);
                if(link_id != -1)
                {   
                    char s_direction[32];
                    char s_type[32];
                    char no_src[256];
                    char no_dst[256];
                    int src_id;
                    int dst_id;
                    int acc_id;

                    cons[id].status.active = 1;
                    cons[id].status.last_menu = cons[id].status.current_menu;
                    cons[id].status.current_menu = menus[link_id];
                    if(cons[id].status.last_menu == NULL)
                        cons[id].status.last_menu = cons[id].status.current_menu;                                        
                    
                    snprintf(s_direction, sizeof(s_direction), "%s", cons[id].s_direction);
                    snprintf(s_type, sizeof(s_type), "%s", cons[id].s_type);
                    snprintf(no_src, sizeof(no_src), "%s", cons[id].no_src);
                    snprintf(no_dst, sizeof(no_dst), "%s", cons[id].no_dst);
                    src_id = cons[id].conid_src;
                    dst_id = cons[id].conid_dst;
                    acc_id = cons[id].acc_id;

                    // send "GOTO" event                                                        
                    plugin_event("", "GOTO", s_direction, s_type, src_id, dst_id, no_src, no_dst, arg[i+1], acc_id+1);
                    ret++;                        
                }                
            }
        }
    }

    // >0: command executed, =0: show help
    return ret;
}
