/*********************************************************************************
 * port_soundcard.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "port_soundcard.c"

pj_status_t port_soundcard_init(connection_t *mCON)
{
    #if DTMFBOX_USE_SOUND

    connection_t *CUR_CON = (connection_t*)mCON;    
    pj_status_t status = PJ_SUCCESS;
    int i;

    PJ_ASSERT_RETURN(CUR_CON != NULL, 1);    
    PJ_ASSERT_RETURN(CUR_CON->ports->custom == NULL, 2);    

    strcpy(oGLB.audio.snd_name_play, "");
    strcpy(oGLB.audio.snd_name_rec, "");
    
    //pj_ansi_snprintf(env, sizeof(env), "PA_MIN_LATENCY_MSEC=0");      // set PortAudio/DirectX latency environment variable
    //putenv(env);

    if(!oSET.audio.null_sound)
    {
        status = pjmedia_snd_init(&oGLB.cp.factory);
        
        if(status == PJ_SUCCESS)
        {
            int dev_cnt = pjmedia_snd_get_dev_count();

            if(dev_cnt)
            {
                app_msg(THIS_FILE, "%d sound device(s) found!", dev_cnt);

                // show sound devices
                for(i=0; i<dev_cnt; i++)
                {
                    const pjmedia_snd_dev_info* snd_info = pjmedia_snd_get_dev_info(i);
                    app_dbg(THIS_FILE, "#%d - %s", i, snd_info->name);
                }
                            
                // Bidirectional or unidirectional sound?
                if(oSET.audio.bidirectional_sound)
                {
                    // Create sound devices bidirectional                    
                    status = pjmedia_snd_port_create(oGLB.pool, oSET.audio.snd_play, oSET.audio.snd_rec, oSET.audio.snd_clockrate, CONF_CHANNELS, oSET.audio.snd_clockrate * 20 / 1000 * CONF_CHANNELS, CONF_BITS, 0, &oGLB.audio.snd_bidi_port);
                    if(status != PJ_SUCCESS)
                    {
                        app_err(THIS_FILE, "(%d) Unable to create sound port for player (%d) and recorder (%d)", status, oSET.audio.snd_play, oSET.audio.snd_rec);                    
                        oSET.audio.null_sound = 1;
                        pjmedia_snd_deinit();
                        return 1;
                    }

                    oGLB.audio.snd_play_port = oGLB.audio.snd_bidi_port;
                    oGLB.audio.snd_rec_port = oGLB.audio.snd_bidi_port;
                }
                else
                {
                    // Create sound devices unidirectional (playing and recording seperate)
                    status = pjmedia_snd_port_create_player(oGLB.pool, oSET.audio.snd_play, oSET.audio.snd_clockrate, CONF_CHANNELS, oSET.audio.snd_clockrate * 20 / 1000 * CONF_CHANNELS, CONF_BITS, 0, &oGLB.audio.snd_play_port);
                    if(status != PJ_SUCCESS)
                    {
                        app_err(THIS_FILE, "(%d) Unable to create sound port for player (%d)", status, oSET.audio.snd_play);                    
                        oSET.audio.null_sound = 1;
                        pjmedia_snd_deinit();
                        return 1;
                    }

                    status = pjmedia_snd_port_create_rec(oGLB.pool, oSET.audio.snd_rec, oSET.audio.snd_clockrate, CONF_CHANNELS, oSET.audio.snd_clockrate * 20 / 1000 * CONF_CHANNELS, CONF_BITS, 0, &oGLB.audio.snd_rec_port);
                    if(status != PJ_SUCCESS)
                    {
                        app_err(THIS_FILE, "(%d) Unable to create sound port recorder (%d)", status, oSET.audio.snd_rec);                    
                        oSET.audio.null_sound = 1;
                        pjmedia_snd_deinit();
                        return 1;
                    }
                }

                
                {                    
                    pjmedia_snd_stream_info si;
                    
                    // Get currently selected sounddevice
                    pjmedia_snd_stream_get_info(pjmedia_snd_port_get_snd_stream(oGLB.audio.snd_play_port), &si);
                    if (si.play_id >= 0) 
                        strncpy(oGLB.audio.snd_name_play, pjmedia_snd_get_dev_info(si.play_id)->name, sizeof(oGLB.audio.snd_name_play));                    
                    app_msg(THIS_FILE, "Playback device: %s (#%d - latency: %dms)", oGLB.audio.snd_name_play, si.play_id, si.play_latency);                    

                    pjmedia_snd_stream_get_info(pjmedia_snd_port_get_snd_stream(oGLB.audio.snd_rec_port), &si);
                    if (si.rec_id >= 0) 
                        strncpy(oGLB.audio.snd_name_rec, pjmedia_snd_get_dev_info(si.rec_id)->name, sizeof(oGLB.audio.snd_name_rec));
                    app_msg(THIS_FILE, "Capture device : %s (#%d - latency: %dms)", oGLB.audio.snd_name_rec, si.rec_id, si.rec_latency);

                    // Set echo canceller (bidirectional mode only!)
                    if(oSET.audio.bidirectional_sound)
                    {   
                        if(oSET.audio.snd_ec_tail)
                        {
                            app_dbg(THIS_FILE, "Enable echo canceller with %dms tail length", oSET.audio.snd_ec_tail);                        
                            status = pjmedia_snd_port_set_ec(oGLB.audio.snd_bidi_port, oGLB.pool, oSET.audio.snd_ec_tail, PJMEDIA_ECHO_SIMPLE);
                        }                        
                        else
                        {
                            app_dbg(THIS_FILE, "Disable echo canceller");                        
                            status = pjmedia_snd_port_set_ec(oGLB.audio.snd_bidi_port, oGLB.pool, 0, 0);
                        }

                        if(status != PJ_SUCCESS)                     
                            app_err(THIS_FILE, "(%d) Cannot add echo canceller to sound port!", status);
                    }                    

                    // Create sound port (splitcomb)                    
                    status = port_soundcard_create(CUR_CON->pools->con_pool, CUR_CON, &CUR_CON->ports->custom);
                    PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);
                }                
            }
            else
            {
                app_err(THIS_FILE, "Unable to find sound device! Using null device.");                
                oSET.audio.null_sound = 1;
                pjmedia_snd_deinit();
            }         
        }       
    }
    
    return status;

    #else
    return 1;
    #endif
}

pj_status_t port_soundcard_close(connection_t *mCON)
{    
    #if DTMFBOX_USE_SOUND

    connection_t *CUR_CON = (connection_t*)mCON;    
    portdata_soundcard *pdata = CUR_CON->ports->custom->port_data.pdata;
    pj_status_t status = PJ_SUCCESS;
      
        if(pdata->capi_port != NULL)
            status = pjmedia_port_destroy(pdata->capi_port);

        if(pdata->resample_port != NULL)
            status = pjmedia_port_destroy(pdata->resample_port);

        if(CUR_CON->ports->custom != NULL)
            status = pjmedia_port_destroy(CUR_CON->ports->custom);
        
        pj_thread_sleep(250);          

        if(oGLB.audio.snd_bidi_port != NULL)        
            status = pjmedia_snd_port_disconnect(oGLB.audio.snd_bidi_port);                

        if(oGLB.audio.snd_play_port != NULL)        
            status = pjmedia_snd_port_disconnect(oGLB.audio.snd_play_port);                
        
        if(oGLB.audio.snd_rec_port != NULL)        
            status = pjmedia_snd_port_disconnect(oGLB.audio.snd_rec_port);                

        pj_thread_sleep(250);          

        if(oGLB.audio.snd_bidi_port != NULL)        
            status = pjmedia_snd_port_destroy(oGLB.audio.snd_bidi_port);

        if(oGLB.audio.snd_play_port != NULL)        
            status = pjmedia_snd_port_destroy(oGLB.audio.snd_play_port);

        if(oGLB.audio.snd_rec_port != NULL)        
            status = pjmedia_snd_port_destroy(oGLB.audio.snd_rec_port);
   
        oGLB.audio.snd_bidi_port = NULL;
        oGLB.audio.snd_play_port = NULL;        
        oGLB.audio.snd_rec_port = NULL;        
        CUR_CON->ports->custom = NULL;
        
    return status;

    #else
    return 0;
    #endif
}

static pj_status_t port_soundcard_create(pj_pool_t *pool, connection_t *mCON, pjmedia_port **p_port)
{   
    #if DTMFBOX_USE_SOUND

    connection_t *CUR_CON = (connection_t*)mCON;    
    pjmedia_port *port;
    portdata_soundcard *pdata;    
    pj_status_t status = PJ_SUCCESS;
    unsigned count;    
    
    PJ_ASSERT_RETURN(pool, PJ_EINVAL);

    port = pj_pool_zalloc(pool, sizeof(pjmedia_port));
    PJ_ASSERT_RETURN(port != NULL, PJ_ENOMEM);

    // Fill port info
    port->info.bits_per_sample = CONF_BITS;
    port->info.channel_count = CONF_CHANNELS;
    port->info.encoding_name = pj_str("pcm");
    port->info.has_info = 1;
    port->info.name = pj_str("soundcard port");
    port->info.need_info = 0;
    port->info.pt = 0xFF;
    port->info.clock_rate = oSET.audio.conf_clockrate;
    port->info.samples_per_frame = (port->info.clock_rate * 20 / 1000); 
    port->info.bytes_per_frame = port->info.samples_per_frame * (CONF_BITS / 8);
    port->info.type = PJMEDIA_TYPE_AUDIO;
    
    // Initialize callbacks
    port->put_frame = &port_soundcard_put_frame;
    port->get_frame = &port_soundcard_get_frame;

    // Create snd_port pdata 
    port->port_data.pdata = pdata = pj_pool_zalloc(pool, sizeof(portdata_soundcard));
    
    // Fill snd_port pdata
    count = port->info.samples_per_frame / CONF_CHANNELS;
    pdata->con_id = CUR_CON->id;        
    pdata->mixdata.tx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_tx_volume)); 
    pdata->mixdata.rx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_rx_volume));
    pdata->mixdata.mother_port = port;
    
    // Handle soundport as connection. This makes things easier.
    CUR_CON->info->type = SCRD;
    CUR_CON->info->active = 1;
    CUR_CON->info->direction = NONE;
    CUR_CON->capi->NCCI = (unsigned long)-1;
    CUR_CON->capi->PLCI = (unsigned long)-1;
    CUR_CON->capi->hdl = 0;    

    // Create (private) CAPI port    
    status = port_capi_create(pool, CUR_CON, &pdata->capi_port);
    pj_assert(status == PJ_SUCCESS);

    // Create (public) tone port
    status = port_tone_create(pool, CUR_CON, &CUR_CON->ports->tone, 0, 0, 0, 0, 0);
    pj_assert(status == PJ_SUCCESS);

    // Create resample port
    if(oSET.audio.snd_clockrate != oSET.audio.conf_clockrate)
    {
        app_msg(THIS_FILE, "Init resampler for soundcard (%dhz -> %dhz)", oSET.audio.snd_clockrate, oSET.audio.conf_clockrate);
        status = pjmedia_resample_port_create(oGLB.pool, port, oSET.audio.snd_clockrate, oSET.audio.resampler_quality, &pdata->resample_port);
        pj_assert(status == PJ_SUCCESS);
    }
    else {        
        pdata->resample_port = port;
    }

    // Connect soundcard with resample port (--> capi_port)
    if(oSET.audio.bidirectional_sound)
    {
        // Bidirectional
        status = pjmedia_snd_port_connect(oGLB.audio.snd_bidi_port, pdata->resample_port);    
        pj_assert(status == PJ_SUCCESS);
    }
    else
    {
        // Unidirectional
        status = pjmedia_snd_port_connect(oGLB.audio.snd_play_port, pdata->resample_port);    
        pj_assert(status == PJ_SUCCESS);

        status = pjmedia_snd_port_connect(oGLB.audio.snd_rec_port, pdata->resample_port);    
        pj_assert(status == PJ_SUCCESS);
    }
    
    *p_port = port;

    return status;

    #else
    return 1;
    #endif
}



static pj_status_t port_soundcard_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{  
    portdata_soundcard *pdata = port->port_data.pdata;    
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;
    pj_status_t status = PJ_SUCCESS;
   
    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;

    // Receive from reverse channel (capi port)    
    status = pjmedia_port_get_frame(pdata->capi_port, &frame_new);
    
    // Important! Set frame type!!
    if(status == PJ_SUCCESS)
        frame->type = PJMEDIA_FRAME_TYPE_AUDIO;
    
    // Mixer callback (not needed, because already done in capi_port)
    //if(port && pdata)    
      //  mixer_port_get_frame(port, &frame_new, pdata->con_id, -1, status);

    return status;
}

static pj_status_t port_soundcard_put_frame(pjmedia_port *port, const pjmedia_frame *frame)
{             
    portdata_soundcard *pdata = port->port_data.pdata;    
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;

    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;

    // Mixer callback (not needed, because already done in capi_port) 
    //mixer_port_put_frame(port, &frame_new, pdata->con_id, -1);
    
    // Send to reverse channel (capi port)        
    return pjmedia_port_put_frame(pdata->capi_port, &frame_new);    
}

