/*********************************************************************************
 * script_funcs.c ($Revision: 192 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "script_funcs.c"

// Execute actions received from second instance
int script_do_actions()
{    
    char name[32];
    int new_con = -1;
    int play_id = -1;
    int rec_id = -1;
    int timer_id = -2;
    int sh_conid = -1;
    int ret_id = -1;
    int pipe_id_global = -1;                
        
    // Open named pipe
    snprintf(name, sizeof(name), "dtmfbox_out_%d", sh_actions->queue.cur_pid);         
    pipe_id_global = pipe_open(name, 0, 0, 1);

    // -----------------------------------------------------------------------
    // ! do script actions !
    // -----------------------------------------------------------------------
    // validate connection id
    if(sh_actions->connection > oSET.max_connections || sh_actions->connection < 0)
        sh_actions->connection = -1;
    
    // connection not valid, when free!
    if(sh_actions->connection != -1) {
        if(oCONS[sh_actions->connection].info->type == FREE)                 
            sh_actions->connection = -1;                               
    }
    sh_conid = sh_actions->connection;

    app_dbg3(THIS_FILE, "[%d] (SCRIPT) - Queue: %d, Pending: %d", sh_conid, sh_actions->queue.cur_queue, (sh_actions->queue.in_queue-1));

    // stop daemon?
    if(sh_actions->stop.option == 3)
    {
        app_msg(THIS_FILE, "[%d] (SCRIPT) - stop daemon (pid: %d)...", sh_conid, getpid());
        signal_handler(9);        
        return -1;
    }

    // Show output (clients, accounts, connections)    
    if(!script_write_output(sh_actions->list, pipe_id_global)) {        
        return -1;
    }
    
    // Make a call?
    new_con = script_make_call();
    
    // Set/unset global environment variable
    script_environment();

    // only when connection-id is valid!
    if(sh_conid != -1)
    {
        if(oCONS[sh_conid].info->type != FREE)
        {
            if(oCONS[sh_conid].info->type != SCRD)
            {            
                switch(sh_actions->hook)
                {
                    // -hook down
                    case 1:
                        script_hook_up();
                        break;

                    // -hook up
                    case 2:
                        script_hook_down();
                        break;
                    
                    // -hook hold
                    case 3:
                        script_hook_hold(1);
                        break;

                    // -hook unhold
                    case 4:
                        script_hook_hold(0);
                        break;

                    // -hook reject
                    case 5: 
                        script_hook_reject();
                        break;

                    // -hook alert
                    case 6:
                        script_hook_alert();
                        break;
                }
            }

            // Script timer
            script_timer();
                         
            // Play/Record/Stop wave/tone - when audio is ready, only.            
            DTMFBOX_LOCK();
            script_stop_play_record(&play_id, &rec_id);
            DTMFBOX_UNLOCK();

            // Play tones from tone-generator - when audio is ready, only.
            script_tone_generator();

            // Send DTMF digits to connection - when audio is ready, only.
            script_send_dtmf();
        
            // Send display text to connection
            script_send_text();

            // Bring two connections in conference or remove them
            script_bridge();

            // Mixer control
            script_mixer_control(play_id, rec_id);
            
            // Change scriptfile or delimiter for current connection
            script_change_script();            
        }
    }
            
    // return new connection id (-call parameter)
    //   or a play_id (-play.. -record... parameters)
    //   or a timer_id (-timer parameter)
    if ((new_con == -1) && (play_id != -1)) { ret_id = play_id; }
    else if ((new_con == -1) && (timer_id != -2)) { ret_id = timer_id; }
    else { ret_id = new_con; }

    // Open output pipe
    {
        if(pipe_id_global != -1)
        {
            // connection-id, play-id, timer-id ...
            if(ret_id != -1)
            {
                char buf[32];                
                pj_ansi_snprintf(buf, sizeof(buf), "%d", ret_id);
                pipe_write(pipe_id_global, buf, strlen(buf));
                plugin_raise_output(buf, sh_actions->queue.cur_pid);
            }

            // environment value ...
            if(strcmp(sh_actions->retval.environment, DEVNULL))
                pipe_write(pipe_id_global, sh_actions->retval.environment, strlen(sh_actions->retval.environment));
            
            // close named pipe
            pipe_close(pipe_id_global);
        }            
    }

    return ret_id;
}

// Play a wave, record a wave and stop playing/recording
//   if player was started, return play_id > -1
//   if recorder was started, return rec_id = 0
void script_stop_play_record(int *play_id, int *rec_id)
{    
    int sh_conid = sh_actions->connection;
    int play_id_stop = -1;
    int rec_id_stop = -1;
    int idle = 0;

    // Reset play_id / rec_id
    *play_id = -1;
    *rec_id = -1;
      
    // Sometimes, connection is not ready at this point (directly after -hook up).
    // Just wait some time...    
    while(oCONS[sh_conid].info->status == 1 && idle < 10)
    {
        pj_thread_sleep(20);
        idle++;
    }
    
    // Play wave-file
    if(strcmp(sh_actions->play_rec.play_file, DEVNULL))
    {    
         if(port_filestream_player_recorder(&oCONS[sh_conid], sh_actions->play_rec.play_file, "", sh_actions->play_rec.stream_clockrate, sh_actions->play_rec.speex, sh_actions->play_rec.stream_buffersize, sh_actions->play_rec.wait_start, sh_actions->play_rec.wait_end, play_id) == PJ_SUCCESS)
         {
             int mixer_play_id = sh_actions->mixer.play_id;
             
             if(mixer_play_id != -1) {
               oCONS[sh_conid].mixer->is_background[mixer_play_id] = 0;
               oCONS[sh_conid].mixer->is_talkover[mixer_play_id] = 0;
               set_talkover_active(sh_conid, mixer_play_id, 0);
             }
             
             app_dbg(THIS_FILE, "[%d] (SCRIPT) [play_id: %d] - playing %s...", sh_conid, mixer_play_id, sh_actions->play_rec.play_file);
             //return_play_id = 1;
         }         
    }

    // Record wave-file
    if(strcmp(sh_actions->play_rec.rec_file, DEVNULL))
    {    
         port_filestream_player_recorder(&oCONS[sh_conid], "", sh_actions->play_rec.rec_file, sh_actions->play_rec.stream_clockrate, sh_actions->play_rec.speex, sh_actions->play_rec.stream_buffersize, sh_actions->play_rec.wait_start, sh_actions->play_rec.wait_end, rec_id);
         app_dbg(THIS_FILE, "[%d] (SCRIPT) - recording %s...", sh_conid, sh_actions->play_rec.rec_file);         
    }

    // Stop playing/recording
    switch(sh_actions->stop.option)
    {
        // Stop playing...
        case 1:            
            play_id_stop = sh_actions->stop.play_id;
            port_filestream_player_recorder(&oCONS[sh_conid], DEVNULL, "", 0, 0, 0, 0, 0, &play_id_stop);
            app_dbg(THIS_FILE, "[%d] (SCRIPT) - [play_id: %d] - playing stopped!", sh_conid, play_id_stop);            
            break;

        // Stop recording...
        case 2:            
            port_filestream_player_recorder(&oCONS[sh_conid], "", DEVNULL, 0, 0, 0, 0, 0, &rec_id_stop);
            app_dbg(THIS_FILE, "[%d] (SCRIPT) - recording stopped!", sh_conid);            
            break;
 
        // Stop tone-generator...
        case 4:                   
            sh_actions->tone.active = 1;
            sh_actions->tone.freq1 = 0;
            sh_actions->tone.freq2 = 0;
            sh_actions->tone.on_msec = 0;
            sh_actions->tone.off_msec = 0;
            sh_actions->tone.volume = 0;
            script_tone_generator();
            sh_actions->tone.active = 0;                
            app_dbg(THIS_FILE, "[%d] (SCRIPT) - tone generator stopped!", sh_conid);
            break;
        
        // Stop waitevent block
        case 5:
            oCONS[sh_conid].script->waitevent = 0;
            break;
    }            
}


// Tone generator
void script_tone_generator()
{    
    portdata_tone *pdata;
    pj_status_t status = PJ_SUCCESS;
    int sh_conid = sh_actions->connection;

    if(sh_actions->tone.active == 1 && sh_conid != -1)
    {
        // not valid?
        if(oCONS[sh_conid].ports->tone == NULL || 
           oCONS[sh_conid].slots->tone == -1   || 
           oCONS[sh_conid].slots->media == -1)
        {
            return;
        }

        // Get private portdata
        pdata = oCONS[sh_conid].ports->tone->port_data.pdata;        
        
        // Fill tone portdata                 
        pdata->freq1 = sh_actions->tone.freq1;
        pdata->freq2 = sh_actions->tone.freq2;
        pdata->on_msec = sh_actions->tone.on_msec;
        pdata->off_msec = sh_actions->tone.off_msec;
        pdata->volume = sh_actions->tone.volume;

        if(pdata->tonegen_port != NULL)
        {
            pjmedia_tonegen_stop(pdata->tonegen_port);

            while(pjmedia_tonegen_is_busy(pdata->tonegen_port))           
                pj_thread_sleep(25);

            pjmedia_tonegen_stop(pdata->tonegen_port);
        }
        
        // disconnect tone_slot from all connections        
        mixer_conf_port_disconnect_all(oCONS[sh_conid].slots->tone);                                    

        if(pdata->freq1 == 0 && pdata->freq2 == 0 && pdata->on_msec == 0 && pdata->off_msec == 0 && pdata->volume == 0)
        {           
        }
        else
        {               
            // reconnect current tone_slot
            status = pjmedia_conf_connect_port(oGLB.conf.bridge, oCONS[sh_conid].slots->tone, oCONS[sh_conid].slots->media, 0);
            if(status == PJ_SUCCESS)
                app_dbg(THIS_FILE, "[%d] (SCRIPT) - playing generated tones... (freq1: %d, freq2: %d, msec on: %d, msec off: %d, volume: %d)",  sh_conid, sh_actions->tone.freq1, sh_actions->tone.freq2, sh_actions->tone.on_msec, sh_actions->tone.off_msec, sh_actions->tone.volume);                    
        }               
     }
}

// Hook down
void script_hook_down()
{
    int sh_conid = sh_actions->connection;

    // VOIP
    if(oCONS[sh_conid].info->type == VOIP)
        voip_hangup(&oCONS[sh_conid]);        

    // CAPI
    else if(oCONS[sh_conid].info->type == CAPI)
       capi_hangup(&oCONS[sh_conid]);
    
    app_dbg(THIS_FILE, "[%d] (SCRIPT) - hook down!", sh_conid);
}

// Hook up
void script_hook_up()
{
    int sh_conid = sh_actions->connection;
    //int in_conf = -1;

    // VOIP
    if(oCONS[sh_conid].info->type == VOIP) 
        voip_confirm_call(&oCONS[sh_conid]);

    // CAPI
    if(oCONS[sh_conid].info->type == CAPI)
        capi_confirm_call(&oCONS[sh_conid]);

    // disconnect all unactive clients
    if(oCONS[sh_conid].info->status == 2)
        voip_registrar_disconnect_clients(&oCONS[sh_conid], 1);
    
    app_dbg(THIS_FILE, "[%d] (SCRIPT) - hook up!", sh_conid);
}

// Hold & retrieve call
void script_hook_hold(int hold)
{
    int sh_conid = sh_actions->connection;
   
    // CAPI
    if(oCONS[sh_conid].info->type == CAPI && oCONS[sh_conid].info->active && (oCONS[sh_conid].capi->SetupInband == !hold))
       capi_hold_call(&oCONS[sh_conid], hold);

}

void script_hook_alert() 
{
    int sh_conid = sh_actions->connection;
    if(oCONS[sh_conid].info->type == CAPI)
        capi_send_alert_req(&oCONS[sh_conid]);
    if(oCONS[sh_conid].info->type == VOIP) 
        oCONS[sh_conid].info->active = 1;

}

void script_hook_reject()
{
    int sh_conid = sh_actions->connection;
    if(oCONS[sh_conid].info->type == CAPI)
        capi_send_connect_resp(&oCONS[sh_conid], 1);
}

// Make a new call (CAPI or VOIP)
int script_make_call()
{
    char no_src[MAX_URI_LENGTH];
    char no_trg[MAX_URI_LENGTH];
    char *scriptfile;
    char *delimiter;
    char delim[2];
    int i;
    int IsCAPI = -1;
    int acc_id = -1;
    int new_con = -1;

    if(strcmp(sh_actions->call.number, DEVNULL) && strcmp(sh_actions->call.target, DEVNULL))    
    {
        // Search for account by Number
        acc_id = account_by_number(sh_actions->call.number);
        i = acc_id;

        // Search for account by Id
        if(acc_id == -1 && isnumeric(sh_actions->call.number))
        {
            int no = atoi(sh_actions->call.number);
            if(no >= 1 && no <= MAX_ACCOUNTS)
            {
                pj_ansi_snprintf(sh_actions->call.number, sizeof(sh_actions->call.number), "%s", oSET.accounts[no-1].number);
                acc_id = no-1;
                i = acc_id;
            }
        }

        // Account found?
        if(acc_id != -1)
        {
            if(oSET.accounts[i].controller.device == CAPI)
                IsCAPI = 1; // CAPI call (DevType: CAPI)
            else
                IsCAPI = 0; // VoIP call (DevType: VOIP)
        }
        else
        {   
            if(sh_actions->call.capi_controller != 0)
                IsCAPI = 1;     // CAPI call (with controller)
            else
                IsCAPI = 0;     // VoIP call (without controller)
        }

        // CAPI call
        if(IsCAPI == 1)
        {
            pj_memcpy(no_trg, sh_actions->call.target, sizeof(no_trg));
            pj_memcpy(no_src, sh_actions->call.number, sizeof(no_src));
            ltrim((char*)&no_trg); rtrim((char*)&no_trg);
            ltrim((char*)&no_src); rtrim((char*)&no_src);

            // Search for free connection
            i = connection_by_free();
            if(i!=-1)
            {
                int controller = 0;

                // Fill connection details (initialize as outgoing call, PLCI identifier: -2)                        
                oCONS[i].capi->PLCI = (unsigned long)-2;            
                oCONS[i].info->type = CAPI;                
                oCONS[i].info->account_id = acc_id;

                // Controller
                if(sh_actions->call.capi_controller != 0)
                    controller = sh_actions->call.capi_controller;
                else
                    if(acc_id != -1)
                        controller = oSET.accounts[acc_id].controller.ctrl_out;
                      
                if(controller == 0) {
                    app_err(THIS_FILE, "No CAPI controller defined!");
                    reset_connection(&oCONS[i]);
                    return -1;
                }

                // Delimiter         
                if(sh_actions->script.delimiter[0] != '-') {
                    if(!strcmp(sh_actions->script.delimiter, "_"))
                        delim[0] = '\0';
                    else
                        delim[0] = sh_actions->script.delimiter[0];
                } else {
                    delim[0] = oSET.delimiter[0];
                }
                delimiter = delim;

                // Scriptfile
                if(strcmp(sh_actions->script.file, DEVNULL))
                    scriptfile = sh_actions->script.file;
                else
                    scriptfile = oSET.scriptfile;
                
                // Conference with another connection?
                if(sh_actions->connection != -1)
                {                      
                    // connect partners (source can be CAPI or VOIP)
                    mixer_conf_array_add(&oCONS[sh_actions->connection], i);
                    mixer_conf_array_add(&oCONS[i], sh_actions->connection);                    
                }

                // Make CAPI-call
                if(capi_call(&oCONS[i], no_src, no_trg, controller, sh_actions->call.anonymous, scriptfile, delimiter)) {
                    new_con = i;                    
                    app_dbg(THIS_FILE, "[%d] (SCRIPT) - CAPI-CALL - %s -> %s [%d]", sh_actions->connection, no_src, no_trg, oCONS[i].capi->controller);
                }
            }
        }

        // VOIP Call
        if(IsCAPI == 0)
        {            
            // Search for free connection
            i = connection_by_free();
            if(i!=-1)
            {
                // Fill connection details            
                oCONS[i].capi->NCCI = (unsigned)-1;
                oCONS[i].info->type = VOIP;
                oCONS[i].info->account_id = acc_id;
                
                pj_memset(no_trg, 0, sizeof(no_trg));
                pj_memset(no_src, 0, sizeof(no_src));

                // Target No
                if(acc_id != -1)
                {
                    // Got host?
                    if(strchr(sh_actions->call.target, '@') != NULL)
                        pj_ansi_snprintf(no_trg, sizeof(no_trg), "<sip:%s>", sh_actions->call.target);

                    else
                        // No host? Add registrar as host?
                        pj_ansi_snprintf(no_trg, sizeof(no_trg), "<sip:%s@%s>", sh_actions->call.target, oSET.accounts[acc_id].registrar);

                    // no_src, with account (add Registrar ID as host)
                    pj_ansi_snprintf(no_src, sizeof(no_src), "<sip:%s>", oSET.accounts[acc_id].id);

                }
                else
                {
                    // Got host?
                    if(strchr(sh_actions->call.target, '@') != NULL)
                        pj_ansi_snprintf(no_trg, sizeof(no_trg), "<sip:%s>", sh_actions->call.target);
                    else
                        // No host? Add global IP as host?
                        pj_ansi_snprintf(no_trg, sizeof(no_trg), "<sip:%s@%s>", sh_actions->call.target, oGLB.ip_addr_global);

                    {                        
                        char host[MAX_URI_LENGTH];
                        char ip[MAX_URI_LENGTH];
                        int port;
                        pj_status_t status;
                        
                        strncpy(host, no_trg, sizeof(host));
                        sip_strip_host(host);
                        
                        // no_src, without account (add global/local IP as host)
                        /*
                        if(!ip_is_public(host)) {
                            pj_ansi_snprintf(no_src, sizeof(no_src), "<sip:%s@%s:%d>", sh_actions->call.number, oGLB.ip_addr_local, oGLB.udp_port_local);
                        } else {
                        */
                            //pj_ansi_snprintf(no_src, sizeof(no_src), "<sip:%s@%s:%d>", sh_actions->call.number, oGLB.ip_addr_global, oGLB.udp_port_global);
                        
                            status = voip_resolve_transport_address(ip, &port);
                            if(status != PJ_SUCCESS)
                            {
                                app_err(THIS_FILE, "(%d) Unable to resolve transport address", status);
                                reset_connection(&oCONS[i]);
                                return -1;
                            }

                            pj_ansi_snprintf(no_src, sizeof(no_src), "<sip:%s@%s:%d>", sh_actions->call.number, ip, port);
                        //}
                        
                    }

                }

                // Conference with another connection?
                if(sh_actions->connection != -1)
                {
                    // connect partners (source can be CAPI or VOIP)
                    mixer_conf_array_add(&oCONS[sh_actions->connection], i);
                    mixer_conf_array_add(&oCONS[i], sh_actions->connection);                    
                }

                // Delimiter         
                if(sh_actions->script.delimiter[0] != '-') {
                    if(!strcmp(sh_actions->script.delimiter, "_"))
                        delim[0] = '\0';
                    else
                        delim[0] = sh_actions->script.delimiter[0];
                } else {
                    delim[0] = oSET.delimiter[0];
                }

                delimiter = delim;

                // Scriptfile
                if(strcmp(sh_actions->script.file, DEVNULL))
                    scriptfile = sh_actions->script.file;
                else
                    scriptfile = oSET.scriptfile;

                // Make VOIP call
                new_con = voip_call(&oCONS[i], no_src, no_trg, sh_actions->call.anonymous, scriptfile, delimiter);
                app_dbg(THIS_FILE, "[%d] (SCRIPT) - VOIP-CALL - %s -> %s", sh_actions->connection, oCONS[i].info->no_source, oCONS[i].info->no_target);
            }
        }
    }

    return new_con;
}

// Bridge calls (mix audio / conference)
// Parameter: [x] -bridge on [y] | [x] -bridge off [y]
void script_bridge()
{   
    pj_status_t status = 1;
    int sh_conid = sh_actions->connection;

    // Connect with conn-id (audio)
    if(sh_actions->bridge.connect != -1 && sh_actions->bridge.connect < oSET.max_connections)
    {        
        status = mixer_conf_port_connect(&oCONS[sh_conid], &oCONS[sh_actions->bridge.connect], 1);        
    }

    // Disconnect conn-id (audio)
    if(sh_actions->bridge.disconnect != -1 && sh_actions->bridge.disconnect < oSET.max_connections)
    {        
        status = mixer_conf_port_disconnect(&oCONS[sh_conid], &oCONS[sh_actions->bridge.disconnect], 1);
    }    
}

// Send DTMF digits to connection
// Parameter: -dtmf [x]
void script_send_dtmf()
{
    int sh_conid = sh_actions->connection;

    // Send DTMF digits (only active connections)
    if(strcmp(sh_actions->dtmf, DEVNULL) && oCONS[sh_conid].info->active == 1)
    {   
        if(oCONS[sh_conid].info->type == SCRD)
        {                
            // do NOT send dtmf digits to soundcard
        }
        else            
        {
            if(oCONS[sh_conid].info->type == VOIP)
                voip_send_dtmf(sh_actions->dtmf[0], sh_conid);   // VoIP
            else                
                capi_send_dtmf(sh_actions->dtmf[0], sh_conid);   // CAPI      
        }
    }    
}

void script_send_text()
{
    int sh_conid = sh_actions->connection;

    // Send DTMF digits (only active connections)
    if(strcmp(sh_actions->text, DEVNULL) && oCONS[sh_conid].capi->PLCI > 0)
    {  
        if(oCONS[sh_conid].info->type != CAPI) {
           app_err(THIS_FILE, "Display messages are currently supported for CAPI only!");
           return;
        }
           
        capi_display_message(&oCONS[sh_conid], sh_actions->text);
    }
}

// Call script timed
void script_timer()
{
    int sh_conid = sh_actions->connection;

    if (sh_actions->timer.timeout > 0 && sh_conid != -1) 
    {
        // start timer
        pj_time_val timeout;                

        // Copy temporary optionstring to connection struct
        strncpy(oCONS[sh_conid].script->timer_optionstring, sh_actions->timer.optionstring, sizeof(oCONS[sh_conid].script->timer_optionstring));
        
        // Copy scriptfile for timer to connection struct (either from -scriptfile parameter, or connection)
        if(strcmp(sh_actions->script.file, DEVNULL))
            strncpy(oCONS[sh_conid].script->timer_scriptfile, sh_actions->script.file, sizeof(oCONS[sh_conid].script->timer_scriptfile));
        else
            strncpy(oCONS[sh_conid].script->timer_scriptfile, oCONS[sh_conid].script->file, sizeof(oCONS[sh_conid].script->timer_scriptfile));

        // init timer entry (entry->id=con_id, entry->user_data=optionstring)        
        pjsip_endpt_cancel_timer(oGLB.endpt, &oCONS[sh_conid].script->timer);
        pj_timer_entry_init(&oCONS[sh_conid].script->timer, sh_conid, &oCONS[sh_conid].script->timer_optionstring, &script_timer_event);
        
        // schedule timer (sec or msec)
        if(sh_actions->timer.mode == 0)
        {
            timeout.sec = sh_actions->timer.timeout;
            timeout.msec = 0;
        }
        else
        {
            timeout.sec = 0;
            timeout.msec = sh_actions->timer.timeout;
        }

        pjsip_endpt_schedule_timer(oGLB.endpt, &oCONS[sh_conid].script->timer, &timeout);        
    } 
    else 
    {
        // stop timer(s)
        if (sh_actions->timer.timeout == -1) {
            pjsip_endpt_cancel_timer(oGLB.endpt, &oCONS[sh_conid].script->timer);
          /*
          // stop specific timer id
          if (isnumeric(sh_actions->timer.optionstring)) {
            //int timer_id = atoi(sh_actions->timer.optionstring);
            //pj_timer_heap_cancel( ... )
            //reset_timer(timer_id,sh_conid);
          }
          // stop all timers for current connection id
          else
          {
            //reset_timer(-1,sh_conid);
          }
          */
        }
    }        
}

// Write outputfile for -list parameter (read by second instance)
int script_write_output(int iShowConnects, int pipe_id_global)
{    
    char line[512];
    char tmp[512];
    char no_src[MAX_URI_LENGTH];
    char no_dst[MAX_URI_LENGTH];
    char status[16];
    int i, j;
            
    if(iShowConnects == 0)
        return -1;

    if(pipe_id_global == -1)
        return -1;

    // Show connections (-list), show all connections (-list all)
    if(iShowConnects == 1 || iShowConnects == 2)
    {
        for(i=0; i<oSET.max_connections; i++)
        {
            if(oCONS[i].info->type != FREE)
            {                    
                pj_memset(line, 0, sizeof(line));
                pj_memset(no_src, 0, sizeof(no_src));
                pj_memset(no_dst, 0, sizeof(no_dst));
                
                // Src-No / Dst-No
                {
                    char host[MAX_URI_LENGTH];
                    char user[MAX_URI_LENGTH];

                    pj_memcpy(user, oCONS[i].info->no_source, sizeof(user));
                    pj_memcpy(host, oCONS[i].info->no_source, sizeof(host));
                    sip_strip_user(user); sip_strip_host(host);
                    if(strlen(host))
                        pj_ansi_snprintf(no_src, sizeof(no_src), "%s@%s", user, host);
                    else
                        pj_ansi_snprintf(no_src, sizeof(no_src), "%s", oCONS[i].info->no_source);

                    pj_memcpy(user, oCONS[i].info->no_target, sizeof(user));
                    pj_memcpy(host, oCONS[i].info->no_target, sizeof(host));
                    sip_strip_user(user); sip_strip_host(host);
                    if(strlen(host))
                        pj_ansi_snprintf(no_dst, sizeof(no_dst), "%s@%s", user, host);
                    else
                        pj_ansi_snprintf(no_dst, sizeof(no_dst), "%s", oCONS[i].info->no_target);
                }

                // Connection-Id
                pj_ansi_snprintf(line, sizeof(line), " (%d)", i);

                // Status
                switch(oCONS[i].info->status)
                {
                    case 1:
                        pj_ansi_snprintf(status, sizeof(status), "CONNECT   ");
                        break;
                    case 2:
                        pj_ansi_snprintf(status, sizeof(status), "CONFIRMED ");
                        break;
                    case 3:
                        pj_ansi_snprintf(status, sizeof(status), "DISCONNECT");
                        break;
                    default:
                        pj_ansi_snprintf(status, sizeof(status), "INACTIVE  ");
                        break;
                }

                if(oCONS[i].info->type == SCRD)
                {
                    // Soundcard (always connection 0)
                    pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                    pj_ansi_snprintf(line, sizeof(line), "%s [SCRD] SOUNDCRD - ACTIVE     -", tmp);
                }
                else
                {
                    pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);

                    // CAPI
                    if(oCONS[i].info->type == CAPI && (oCONS[i].info->direction == INCOMING || oCONS[i].info->direction == OUTGOING))
                        pj_ansi_snprintf(line, sizeof(line), "%s [CAPI]", tmp);

                    // VOIP
                    if(oCONS[i].info->type == VOIP && (oCONS[i].info->direction == INCOMING || oCONS[i].info->direction == OUTGOING))
                        pj_ansi_snprintf(line, sizeof(line), "%s [VOIP]", tmp);

                    // USER
                    if(oCONS[i].info->type == VOIP && (oCONS[i].info->direction & CLIENT))
                        pj_ansi_snprintf(line, sizeof(line), "%s [USER]", tmp);

                    // INCOMING
                    if(oCONS[i].info->direction & INCOMING) {
                        
                        // got accound id? set number from settings...
                        if(oCONS[i].info->account_id != -1)
                            pj_ansi_snprintf(no_src, sizeof(no_src), "%s", oSET.accounts[oCONS[i].info->account_id].number);

                        sip_strip_user(no_src);
                        pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                        pj_ansi_snprintf(line, sizeof(line), "%s INCOMING - %s - %s <- %s", tmp, status, no_src, no_dst);
                    }

                    // OUTGOING
                    else if(oCONS[i].info->direction & OUTGOING) 
                    {
                        // got accound id? set number from settings...
                        if(oCONS[i].info->account_id != -1)
                            pj_ansi_snprintf(no_src, sizeof(no_src), "%s", oSET.accounts[oCONS[i].info->account_id].number);

                        sip_strip_user(no_src);
                        pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                        pj_ansi_snprintf(line, sizeof(line), "%s OUTGOING - %s - %s -> %s", tmp, status, no_src, no_dst);
                    }
                  
                }

                // Show "connected with"
                // walkthrough conference partners and show connection id
                if(oCONS[i].in_conf_cnt > 0 && strlen(line)) 
                {
                    int found = 0;
                    pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                    pj_ansi_snprintf(line, sizeof(line), "%s (", tmp);

                    for(j=0; j<INCONF_PORTS; j++) {
                        if(oCONS[i].in_conf[j] != -1) {
                            pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                            pj_ansi_snprintf(line, sizeof(line), "%s%d,", tmp, oCONS[i].in_conf[j]);
                            found = 1;
                        }
                    }

                    if(!found) {
                        pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                        pj_ansi_snprintf(line, sizeof(line), "%s- ", tmp);
                    }

                    line[strlen(line) - 1] = '\0';
                    pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                    pj_ansi_snprintf(line, sizeof(line), "%s)\n", tmp);
                }
                else
                {
                    if(strlen(line)) {
                        pj_ansi_snprintf(tmp, sizeof(tmp), "%s", line);
                        pj_ansi_snprintf(line, sizeof(line), "%s\n", tmp);
                    }
                }

                // write line to temporary file
                //fwrite(line, strlen(line), 1, fp);                           
                pipe_write(pipe_id_global, line, strlen(line));
                plugin_raise_output(line, sh_actions->queue.cur_pid);
            }
        }

        pipe_close(pipe_id_global);  
    }

    // Output all Clients (-list clients)
    if(iShowConnects == 3)
    {
        for(i=0; i<oSET.voip.max_clients; i++)
        {
            int j;
            int accounts = 0;

            // how many accounts does the client have?
            for(j=0; j<MAX_ACCOUNTS; j++)
                if(oGLB.clients[i].account_id[j] != -1)
                    accounts++;

            if(accounts != 0)
            {
                snprintf(line, sizeof(line), " (%d) %-45s - %d account(s), %d min\n", i+1, oGLB.clients[i].addr, accounts, oGLB.clients[i].timeout);
                pipe_write(pipe_id_global, line, strlen(line));
                //fprintf(fp, " (%d) %-45s - %d account(s), %d min\n", i+1, oGLB.clients[i].addr, accounts, oGLB.clients[i].timeout);
                plugin_raise_output(line, sh_actions->queue.cur_pid);
            }
        }
        pipe_close(pipe_id_global);
    }

    // Output all accounts (-list accounts)
    if(iShowConnects == 4)
    {        
        for(i=0; i<MAX_ACCOUNTS; i++)
        {
            char type[10];

            if(oSET.accounts[i].active)
            {
                if(oSET.accounts[i].controller.device != CAPI)
                    strcpy(type, "VOIP");
                else
                    strcpy(type, "CAPI");

                switch(oSET.accounts[i].status)
                {
                    case 0:
                        snprintf(line, sizeof(line), " (%d) [%s] %-15s - %-15s - Pending!\n", i+1, type, oSET.accounts[i].number, oSET.accounts[i].registrar);                        
                        //fprintf(fp, " (%d) [%s] %-15s - %-15s - Pending!\n", i+1, type, oSET.accounts[i].number, oSET.accounts[i].registrar);
                        break;
                    case 1:
                        snprintf(line, sizeof(line), " (%d) [%s] %-15s - %-15s - %s (%d sec)\n", i+1, type, oSET.accounts[i].number, oSET.accounts[i].registrar, oSET.accounts[i].status_reason, oSET.accounts[i].status_nextreg);                        
                        //fprintf(fp, " (%d) [%s] %-15s - %-15s - %s (%d sec)\n", i+1, type, oSET.accounts[i].number, oSET.accounts[i].registrar, oSET.accounts[i].status_reason, oSET.accounts[i].status_nextreg);
                        break;
                    case 2:
                        snprintf(line, sizeof(line), " (%d) [%s] %-15s - %-15s - %s\n", i+1, type, oSET.accounts[i].number, oSET.accounts[i].registrar, oSET.accounts[i].status_reason);                        
                        //fprintf(fp, " (%d) [%s] %-15s - %-15s - %s\n", i+1, type, oSET.accounts[i].number, oSET.accounts[i].registrar, oSET.accounts[i].status_reason);
                        break;
                }

                pipe_write(pipe_id_global, line, strlen(line));
                plugin_raise_output(line, sh_actions->queue.cur_pid);
            }
        }
        pipe_close(pipe_id_global);
    }

    // Output all plugins / plugin-help (-list plugins)
    if(iShowConnects == 5)
    {        
        plugin_raise_extra_help(pipe_id_global);
        pipe_close(pipe_id_global);
    }

    return 0;
}

// script timer callback
void script_timer_event(pj_timer_heap_t *timer_heap, struct pj_timer_entry *entry) 
{            
    char *options = (char*)entry->user_data;
    int con_id = entry->id;
    
    PJ_UNUSED_ARG(timer_heap);
    
    script_event_add(oCONS[con_id].info->type, TIMER, oCONS[con_id].script->timer_scriptfile, con_id, oCONS[con_id].in_conf[0], oCONS[con_id].info->no_source, oCONS[con_id].info->no_target, options);
}

// Set/unset global environment variables (available in all script events!)
void script_environment()
{
    // Set global environment variable 
    if(strcmp(sh_actions->script.env_key, DEVNULL))
    {
        char *old_val = getenv(sh_actions->script.env_key);        
        
        if(strcmp(sh_actions->script.env_val, DEVNULL))
        {        
            // set environment variable
            char env[sizeof(sh_actions->script.env_key) + sizeof(sh_actions->script.env_val) + 1];
            pj_ansi_snprintf(env, sizeof(env), "%s=%s", sh_actions->script.env_key, sh_actions->script.env_val);                
            app_dbg(THIS_FILE, "[%d] (SCRIPT) - set environment variable.\n old value: \"%s=%s\"\n new value: \"%s=%s\"", sh_actions->connection, sh_actions->script.env_key, old_val, sh_actions->script.env_key, sh_actions->script.env_val);

            #if DTMFBOX_WIN32
                putenv(env);
            #else
                unsetenv(sh_actions->script.env_key);
                setenv(sh_actions->script.env_key, sh_actions->script.env_val, 1);                                
            #endif

            strncpy(sh_actions->script.env_val, DEVNULL, sizeof(sh_actions->script.env_val));                        
        }
        else
        {   
            // return environment variable
            if(old_val != NULL) {
                strncpy(sh_actions->retval.environment, old_val, sizeof(sh_actions->retval.environment));
                app_dbg(THIS_FILE, "[%d] (SCRIPT) - get environment variable: \"%s=%s\"", sh_actions->connection, sh_actions->script.env_key, sh_actions->retval.environment);
            }

            /*
            // unset environment variable
            #if DTMFBOX_WIN32
                char env[sizeof(sh_actions->script.env_key) + sizeof(sh_actions->script.env_val) + 1];
                pj_ansi_snprintf(env, sizeof(env), "%s=", sh_actions->script.env_key);                
                putenv(env);
            #else                                
                unsetenv(sh_actions->script.env_key);
            #endif           
            */
        }        
    }
    else
    {
        strncpy(sh_actions->script.env_val, DEVNULL, sizeof(sh_actions->script.env_val));                        
    }
}

// Change scriptfile or delimiter for a specified connection
void script_change_script()
{
    int sh_conid = sh_actions->connection;

    // do not change scriptfile, when script-timer is selected!!!
    // use scriptfile parameter for timer!
    if(sh_actions->timer.timeout > 0)
        return;

    // Change scriptfile
    if(strcmp(sh_actions->script.file, DEVNULL))
    {
        strncpy(oCONS[sh_conid].script->file, sh_actions->script.file, sizeof(oCONS[sh_conid].script->file));                
        app_dbg(THIS_FILE, "[%d] (SCRIPT) - change scriptfile to \"%s\"", sh_conid, oCONS[sh_conid].script->file);
    }

    // Change delimiter
    if(strcmp(sh_actions->script.delimiter, "-"))
    {
        if(!strcmp(sh_actions->script.delimiter, "_"))
            oCONS[sh_conid].script->delimiter[0] = '\0';
        else
            strncpy(oCONS[sh_conid].script->delimiter, sh_actions->script.delimiter, sizeof(oCONS[sh_conid].script->delimiter));

        // Erase current DTMF buffer
        pj_memset(oCONS[sh_conid].script->dtmf, 0, sizeof(oCONS[sh_conid].script->dtmf));
        oCONS[sh_conid].script->dtmf_pos = 0;

        app_dbg(THIS_FILE, "[%d] (SCRIPT) - change delimiter to \"%s\"", sh_conid, oCONS[sh_conid].script->delimiter);
    }

    // WaitEvent (block script)?
    if(sh_actions->script.waitevent) 
    {
        char name[32];

        // close any existing pipe
        if(oCONS[sh_conid].script->pipe_id != -1) 
            pipe_close(oCONS[sh_conid].script->pipe_id);
                    
        // open new pipe
        snprintf(name, sizeof(name), "dtmfbox_waitevent_%d", sh_actions->queue.cur_pid);
        oCONS[sh_conid].script->pipe_id = pipe_open(name, 0, 0, 1);

        // waitevent type
        if(oCONS[sh_conid].script->pipe_id != -1) {
            oCONS[sh_conid].script->waitevent = sh_actions->script.waitevent;
            app_dbg(THIS_FILE, "[%d] (SCRIPT) - wait for event (%d)", sh_conid, sh_actions->script.waitevent);
        }
    }
}
int script_event_add(connection_devtype_e type, enum SCRIPT_EVENTS m_event, char *scriptfile, int src_id, int dst_id, char *source_no, char *target_no, char *reserved)
{
    int i;
    int store_event = -1;
    int tries = 10;

retry:

    DTMFBOX_LOCK();

    for(i=0; i<MAX_SCRIPT_EVENTS; i++)
    {
        if(script_events[i].type == FREE)
        {
            script_events[i].type = type;
            script_events[i].s_event = m_event;
            script_events[i].conid_src = src_id;
            script_events[i].conid_dst = dst_id;
            script_events[i].direction = oCONS[src_id].info->direction;
            script_events[i].acc_id = oCONS[src_id].info->account_id;
            script_events[i].waitevent = oCONS[src_id].script->waitevent;
            pj_ansi_snprintf(script_events[i].scriptfile, sizeof(script_events[i].scriptfile), "%s", scriptfile);
            pj_ansi_snprintf(script_events[i].no_src, sizeof(script_events[i].no_src), "%s", source_no);
            pj_ansi_snprintf(script_events[i].no_dst, sizeof(script_events[i].no_dst), "%s", target_no);
            pj_ansi_snprintf(script_events[i].reserved, sizeof(script_events[i].reserved), "%s", reserved);
            
            // Which event?
            switch(script_events[i].s_event)
            {
                case CONNECT:
                    oCONS[src_id].info->status = 1;
                    break;

                case DDI:
                    oCONS[src_id].info->status = 1;
                    break;

                case EARLY:
                    oCONS[src_id].info->status = 1;
                    break;

                case CONFIRMED:
                    oCONS[src_id].info->status = 2;
                    break;

                case DTMF:                
                    oCONS[src_id].info->status = 2;
                    break;

                case UNCONFIRMED:
                    oCONS[src_id].info->status = 2;
                    break;

                case DISCONNECT:
                    oCONS[src_id].info->status = 3;
                    break;
            }

            store_event = i;
            break;
        }
    }
    DTMFBOX_UNLOCK();

    if(store_event == -1 && tries >= 0) {
        app_err(THIS_FILE, "Too many script events in queue... wait (%d)!", tries);                
        pj_thread_sleep(5);
        tries--;
        goto retry;
    }

    if(store_event == -1 && tries < 0) {
        app_err(THIS_FILE, "Too many script events in queue... failure!");
    }

    return store_event;
}

int script_event_run()
{
    script_event_t m_script_events[MAX_SCRIPT_EVENTS];
    char scriptfile_quoted[512];    
    char reserved_string[512];
    char dst_no[MAX_URI_LENGTH];
    char src_no[MAX_URI_LENGTH];
    char in_out_str[12];
    char is_voip_str[5];
    char s_event[16];
    char s_con_src[8];
    char s_con_dst[8];
    char s_acc_id[8];
    int in_out = 0;
    int acc_id = -1;
    int i;
    /*
#if (DTMFBOX_WIN32 == 0)
    int fork_pid = 0;
    int fork_status = 0;
#endif
    */

    // No events in queue? Abort!
    if(script_events[0].type == FREE)
        return 0;

    DTMFBOX_LOCK();
    pj_memcpy(m_script_events, script_events, sizeof(m_script_events));        
    pj_memset(script_events, 0, sizeof(script_events));
    DTMFBOX_UNLOCK();

    for(i=0; i<MAX_SCRIPT_EVENTS; i++)
    {
        // When event type is "FREE", the next events are also...
        if(m_script_events[i].type == FREE)
            return 0;
        
        if(m_script_events[i].type != FREE)
        {
            int iCONsrc = m_script_events[i].conid_src;
            int iCONdst = m_script_events[i].conid_dst;

            // Copy numbers 
            pj_ansi_snprintf(src_no, sizeof(src_no), "%s", m_script_events[i].no_src);
            pj_ansi_snprintf(dst_no, sizeof(dst_no), "%s", m_script_events[i].no_dst);            
            ltrim(src_no); rtrim(src_no);
            ltrim(dst_no); rtrim(dst_no);

            // INCOMING/OUTGOING?
            if(m_script_events[i].direction  & INCOMING)
                in_out = 1;
            else 
                in_out = 2;
                                    
            // Incoming/Outgoing?
            if(in_out == 1)       
                strcpy(in_out_str, "INCOMING");
            if(in_out == 2) 
                strcpy(in_out_str, "OUTGOING");
            
            // VoIP, CAPI, SCRD or USER (registrar client)? 
            if(m_script_events[i].type == VOIP) {
                strcpy(is_voip_str, "VOIP"); 
            } else if(m_script_events[i].type == CAPI) {
                strcpy(is_voip_str, "CAPI");
            } else {
                strcpy(is_voip_str, "SCRD");
            }

            if(m_script_events[i].direction & CLIENT)    
                strcpy(is_voip_str, "USER");

            // When account id is available, set configurated number as account no.
            if(m_script_events[i].acc_id >= 0)
            {
                acc_id = m_script_events[i].acc_id + 1;
                pj_memset(src_no, 0, sizeof(src_no));        
                pj_memcpy(src_no, oSET.accounts[acc_id-1].number, sizeof(src_no));
            }    
            sip_strip_user(src_no);
                
            // strip "sip:" and ":port" from target number (VoIP only)
            if(m_script_events[i].type == VOIP)
            {
                char user[MAX_URI_LENGTH];
                char host[MAX_URI_LENGTH];

                pj_memset(user, 0, sizeof(user));
                pj_memset(host, 0, sizeof(host));
                pj_memcpy(user, dst_no, sizeof(user));
                pj_memcpy(host, dst_no, sizeof(host));
                sip_strip_user(user);
                sip_strip_host(host);
                if(strlen(host))
                    pj_ansi_snprintf(dst_no, sizeof(dst_no), "%s@%s", user, host);
                else
                    pj_ansi_snprintf(dst_no, sizeof(dst_no), "%s", user);
            }

            // Which event?
            switch(m_script_events[i].s_event)
            {
                case CONNECT:                    
                    pj_ansi_snprintf(s_event, sizeof(s_event), "CONNECT");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
                    break;

                case DDI:                    
                    pj_ansi_snprintf(s_event, sizeof(s_event), "DDI");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "%s", m_script_events[i].reserved);            
                    break;

                case EARLY:                    
                    pj_ansi_snprintf(s_event, sizeof(s_event), "EARLY");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
                    break;

                case CONFIRMED:                    
                    pj_ansi_snprintf(s_event, sizeof(s_event), "CONFIRMED");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
                    break;

                case DTMF:                    
                    pj_ansi_snprintf(s_event, sizeof(s_event), "DTMF");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "%s", m_script_events[i].reserved);            
                    break;

                case UNCONFIRMED:                    
                    pj_ansi_snprintf(s_event, sizeof(s_event), "UNCONFIRMED");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
                    break;

                case DISCONNECT:                    
                    pj_ansi_snprintf(s_event, sizeof(s_event), "DISCONNECT");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), m_script_events[i].reserved);            
                    break;

                case TIMER:            
                    pj_ansi_snprintf(s_event, sizeof(s_event), "TIMER");
                    pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "%s", m_script_events[i].reserved);            
                    break;
            }

            // Raise event in scriptfile
            if(strlen(m_script_events[i].scriptfile))
            {
                // convert some other values to char (for execl/spawnl)
                pj_ansi_snprintf(s_con_src, sizeof(s_con_src), "%d", iCONsrc);
                pj_ansi_snprintf(s_con_dst, sizeof(s_con_dst), "%d", iCONdst);
                pj_ansi_snprintf(s_acc_id, sizeof(s_acc_id), "%d", acc_id);        
                pj_ansi_snprintf(scriptfile_quoted, sizeof(scriptfile_quoted), "\"%s\"", m_script_events[i].scriptfile);
                
                // spawn a new child process in background        
                if(m_script_events[i].waitevent == 0)
                {
                    app_msg(THIS_FILE, "RUN SCRIPT: %s \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"", scriptfile_quoted, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string);
                    
                    #if DTMFBOX_WIN32     
                    {   
                        spawnl(P_NOWAITO, m_script_events[i].scriptfile, scriptfile_quoted, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string, NULL);
                    }        
                    #else                
                    {               
                        // use double fork() trick to spawn a process that runs the script             
                        int pid,pid2,status,died;            
                        if (pid=fork()) {
                          died = waitpid(pid, &status, 0); //,WNOHANG);
                        } else {
                          // drop priority for child processes
                          setpriority(PRIO_PROCESS, 0, +9);
                          if (pid2 = fork()) {
                            exit(0);
                          } 
                          else 
                          {
                            if(m_script_events[i].s_event != TIMER)
                            {
                                // invoke shell with execl
                                execl(m_script_events[i].scriptfile, scriptfile_quoted, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string, NULL);                                
                            }
                            else
                            {
                                // invoke shell with system (TIMER only, because of the optional command line arguments!)
                                char cmd[2048];
                                pj_ansi_snprintf(cmd, sizeof(cmd), "%s  \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" %s &", m_script_events[i].scriptfile, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string);
                                system(cmd);              
                            }

                            // obsolete: raise priority for ourselves again
                            //                setpriority(PRIO_PROCESS,0,-9);
                            exit(0);
                          }
                        }               
                    }                
                    #endif      
                }
                else
                {   
                    char events[2048];
                    int pipe_id = oCONS[iCONsrc].script->pipe_id;
                    pj_ansi_snprintf(events, sizeof(events), "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n", s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string);

                    // write to pipe
                    if(pipe_id != -1) {
                        pipe_write(pipe_id, events, strlen(events));                
                    }

                    // capture only one event?
                    if(m_script_events[i].waitevent == 1 || m_script_events[i].s_event == DISCONNECT) {
                        pipe_close(pipe_id);
                        oCONS[iCONsrc].script->pipe_id = -1;
                        oCONS[iCONsrc].script->waitevent = 0;
                    }
                }
            }
            else
            {
                app_err(THIS_FILE, "Unable to run script! No scriptfile defined.");
            }

            // Raise event in plugins
            plugin_raise_event(m_script_events[i], s_event, in_out_str, is_voip_str, src_no, dst_no, reserved_string);
        }
    }

    return 0;
}
/*
// runs scriptfile (system exec)
int script_event(connection_devtype_e IsVoip, enum SCRIPT_EVENTS oEvent, char *scriptfile, int iCONsrc, int iCONdst, char *NoSrc, char *NoDst, char *ReservedString)
{    
    char scriptfile_quoted[512];    
    char reserved_string[512];
    char dst_no[MAX_URI_LENGTH];
    char src_no[MAX_URI_LENGTH];
    char in_out_str[12];
    char is_voip_str[5];
    char s_event[16];
    char s_con_src[8];
    char s_con_dst[8];
    char s_acc_id[8];
    int in_out = 0;
    int acc_id = -1;
    
    // Copy numbers 
    pj_memset(src_no, 0, sizeof(src_no));
    pj_memset(dst_no, 0, sizeof(dst_no));        
    strncpy(src_no, NoSrc, strlen(NoSrc));
    strncpy(dst_no, NoDst, strlen(NoDst));
    ltrim(src_no); rtrim(src_no);
    ltrim(dst_no); rtrim(dst_no);

    // INCOMING/OUTGOING?
    if(oCONS[iCONsrc].info->direction & INCOMING)
        in_out = 1;
    else 
        in_out = 2;
                            
    // Incoming/Outgoing?
    if(in_out == 1)       
        strcpy(in_out_str, "INCOMING");
    if(in_out == 2) 
        strcpy(in_out_str, "OUTGOING");
    
    // VoIP, CAPI, SCRD or USER (registrar client)? 
    if(IsVoip == VOIP) {
        strcpy(is_voip_str, "VOIP"); 
    } else if(IsVoip == CAPI) {
        strcpy(is_voip_str, "CAPI");
    } else {
        strcpy(is_voip_str, "SCRD");
    }

    if(oCONS[iCONsrc].info->direction & CLIENT)    
        strcpy(is_voip_str, "USER");

    // When account id is available, set configurated number as account no.
    if(oCONS[iCONsrc].info->account_id >= 0)
    {
        acc_id = oCONS[iCONsrc].info->account_id + 1;
        pj_memset(src_no, 0, sizeof(src_no));        
        pj_memcpy(src_no, oSET.accounts[acc_id-1].number, sizeof(src_no));
    }    
    sip_strip_user(src_no);
        
    // strip "sip:" and ":port" from target number (VoIP only)
    if(IsVoip == VOIP)
    {
        char user[MAX_URI_LENGTH];
        char host[MAX_URI_LENGTH];

        pj_memset(user, 0, sizeof(user));
        pj_memset(host, 0, sizeof(host));
        pj_memcpy(user, dst_no, sizeof(user));
        pj_memcpy(host, dst_no, sizeof(host));
        sip_strip_user(user);
        sip_strip_host(host);
        if(strlen(host))
            pj_ansi_snprintf(dst_no, sizeof(dst_no), "%s@%s", user, host);
        else
            pj_ansi_snprintf(dst_no, sizeof(dst_no), "%s", user);
    }

    // Which event?
    switch(oEvent)
    {
        case CONNECT:
            oCONS[iCONsrc].info->status = 1;
            pj_ansi_snprintf(s_event, sizeof(s_event), "CONNECT");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
            break;

        case DDI:
            oCONS[iCONsrc].info->status = 1;
            pj_ansi_snprintf(s_event, sizeof(s_event), "DDI");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "%s", oCONS[iCONsrc].script->dtmf);            
            break;

        case EARLY:
            oCONS[iCONsrc].info->status = 1;
            pj_ansi_snprintf(s_event, sizeof(s_event), "EARLY");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
            break;

        case CONFIRMED:
            oCONS[iCONsrc].info->status = 2;
            pj_ansi_snprintf(s_event, sizeof(s_event), "CONFIRMED");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
            break;

        case DTMF:                
            oCONS[iCONsrc].info->status = 2;
            pj_ansi_snprintf(s_event, sizeof(s_event), "DTMF");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "%s", oCONS[iCONsrc].script->dtmf);            
            break;

        case UNCONFIRMED:
            oCONS[iCONsrc].info->status = 2;
            pj_ansi_snprintf(s_event, sizeof(s_event), "UNCONFIRMED");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
            break;

        case DISCONNECT:
            oCONS[iCONsrc].info->status = 3;
            pj_ansi_snprintf(s_event, sizeof(s_event), "DISCONNECT");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "");            
            break;

        case TIMER:            
            pj_ansi_snprintf(s_event, sizeof(s_event), "TIMER");
            pj_ansi_snprintf(reserved_string, sizeof(reserved_string), "%s", ReservedString);            
            break;
    }

    if(strlen(scriptfile))
    {
        // convert some other values to char (for execl/spawnl)
        pj_ansi_snprintf(s_con_src, sizeof(s_con_src), "%d", iCONsrc);
        pj_ansi_snprintf(s_con_dst, sizeof(s_con_dst), "%d", iCONdst);
        pj_ansi_snprintf(s_acc_id, sizeof(s_acc_id), "%d", acc_id);        
        pj_ansi_snprintf(scriptfile_quoted, sizeof(scriptfile_quoted), "\"%s\"", scriptfile);
        
        // spawn a new child process in background        
        if(oCONS[iCONsrc].script->waitevent == 0)
        {
            app_msg(THIS_FILE, "RUN SCRIPT: %s \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"", scriptfile_quoted, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string);
            
            #if DTMFBOX_WIN32     
            {
                spawnl(P_NOWAITO, scriptfile, scriptfile_quoted, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string, NULL);
            }        
            #else                
            {               
                // use double fork() trick to spawn a process that runs the script             
                int pid,pid2,status,died;            
                if (pid=fork()) {
                  died = waitpid(pid, &status, 0); //,WNOHANG);
                } else {
                  // drop priority for child processes
                  setpriority(PRIO_PROCESS, 0, +9);
                  if (pid2 = fork()) {
                    exit(0);
                  } 
                  else 
                  {
                    if(oEvent != TIMER)
                    {
                        // invoke shell with execl
                        execl(scriptfile, scriptfile_quoted, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string, NULL);
                    }
                    else
                    {
                        // invoke shell with system (TIMER only, because of the optional command line arguments!)
                        char cmd[2048];
                        pj_ansi_snprintf(cmd, sizeof(cmd), "%s  \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" %s &", scriptfile, s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string);
                        system(cmd);              
                    }

                    // obsolete: raise priority for ourselves again
                    //                setpriority(PRIO_PROCESS,0,-9);
                    exit(0);
                  }
                }               
            }                
            #endif      
        }
        else
        {   
            char events[2048];
            int pipe_id = oCONS[iCONsrc].script->pipe_id ;
            pj_ansi_snprintf(events, sizeof(events), "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n", s_event, is_voip_str, in_out_str, s_con_src, s_con_dst, src_no, dst_no, s_acc_id, reserved_string);

            // write to pipe
            if(pipe_id != -1) {
                pipe_write(pipe_id, events, strlen(events));                
            }

            // capture only one event?
            if(oCONS[iCONsrc].script->waitevent == 1 || oEvent == DISCONNECT) {
                pipe_close(pipe_id);
                oCONS[iCONsrc].script->pipe_id = -1;
                oCONS[iCONsrc].script->waitevent = 0;
            }
        }
    }
    else
    {
        app_err(THIS_FILE, "Unable to run script! No scriptfile defined.");
    }

    return 0;
}
*/

// reset actions
void reset_actions()
{
    if(sh_actions != NULL)
    {   
        sh_actions->hook = 0;        
        sh_actions->list = 0;
        sh_actions->connection = -1;
        sh_actions->valid_commands = 0;
        //sh_actions->parse_plugin = 0;
        sh_actions->stop.option = 0;
        sh_actions->bridge.connect = -1;
        sh_actions->bridge.disconnect = -1;
        sh_actions->call.capi_controller = 0;
        sh_actions->call.anonymous = 0;
        sh_actions->play_rec.rtp = 0;
        sh_actions->play_rec.wave = 0;        
        sh_actions->play_rec.raw = 0;
        sh_actions->play_rec.speex = 0;
        sh_actions->play_rec.rtp_port = 0;        
        sh_actions->play_rec.stream_clockrate = 0; 
        sh_actions->play_rec.stream_buffersize = 0;
        sh_actions->play_rec.wait_start = 0;
        sh_actions->play_rec.wait_end = 0;        
        sh_actions->timer.mode = 0;
        sh_actions->timer.timeout = 0;        
        sh_actions->mixer.fade_volume = -1;
        sh_actions->mixer.fade_duration = -1;
        sh_actions->mixer.volume = -1;
        sh_actions->mixer.options = 0;
        sh_actions->mixer.play_id = -1;            
        sh_actions->mixer.rec_id = -1;        
        sh_actions->timer.timeout = 0;
        sh_actions->tone.active = 0;
        sh_actions->tone.freq1 = 0;
        sh_actions->tone.freq2 = 0;
        sh_actions->tone.on_msec = 0;
        sh_actions->tone.off_msec = 0;
        sh_actions->tone.volume = 0;             
        sh_actions->script.waitevent = 0;                            
        strcpy(sh_actions->script.env_key, DEVNULL);        
        strcpy(sh_actions->script.file, DEVNULL);
        strcpy(sh_actions->script.delimiter,  "-");
        strcpy(sh_actions->play_rec.play_file, DEVNULL);
        strcpy(sh_actions->play_rec.rec_file, DEVNULL);
        strcpy(sh_actions->call.number, DEVNULL);
        strcpy(sh_actions->call.target, DEVNULL);
        strcpy(sh_actions->dtmf, DEVNULL);         
        strcpy(sh_actions->text, DEVNULL);         
        strcpy(sh_actions->timer.optionstring, "");
        strcpy(sh_actions->retval.environment, DEVNULL);
    }
}

// Create shared memory variable (sh_actions_shmem)
int create_shared_memory()
{
     #if DTMFBOX_WIN32

        // create file mapping for shared memory
        remove(DTMFBOX_SCRIPT_IN);
        shm_id = CreateFile(DTMFBOX_SCRIPT_IN, FILE_READ_DATA | FILE_WRITE_DATA, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_HIDDEN | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
        if(shm_id == NULL || shm_id == (HANDLE)-1)
        {
            app_err(THIS_FILE, "Cannot create shared memory file!");
            return 1;
        }
        shm_map = CreateFileMapping(shm_id, NULL, PAGE_READWRITE, 0, sizeof(struct sh_actions_t), DTMFBOX_SCRIPT_IN);
        CloseHandle(shm_id);

        // Get shared variable
        sh_actions_shmem = (struct sh_actions_t*)MapViewOfFile(shm_map, FILE_MAP_READ | FILE_MAP_WRITE, 0, 0, 0);

    #else

        // allocate shared memory
        shm_id = shmget(DTMFBOX_SCRIPT_IN, sizeof(struct sh_actions_t), IPC_CREAT | 0666);
        if(shm_id == -1) 
        {
           app_err(THIS_FILE, "Cannot allocate shared memory!");
           return 1;           
        }

        // Get shared variable
        sh_actions_shmem = shmat(shm_id, 0, 0);
        if(!sh_actions_shmem)
        {
            app_err(THIS_FILE, "Cannot allocate memory for shared variable!");
            return 1;
        }

    #endif

    return (sh_actions_shmem == NULL);
}

// Open shared memory variable (sh_actions_shmem)
int open_shared_memory()
{   
    if(sh_actions_shmem == NULL)
    {

        // Allocate shared memory (actions) - Windows:
        #if DTMFBOX_WIN32

            // Open already created tmp-file
            shm_map = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, DTMFBOX_SCRIPT_IN);
            if(shm_map != NULL && shm_map != (HANDLE)-1)
            {
                sh_actions_shmem = (struct sh_actions_t*)MapViewOfFile(shm_map, FILE_MAP_READ | FILE_MAP_WRITE , 0, 0, 0);
                if(sh_actions_shmem == NULL)
                {
                    app_err(THIS_FILE, "Cannot allocate shared memory!");
                    return 1;
                }
            }
            else
            {
                app_err(THIS_FILE, "Cannot map shared memory file!");
                return 1;
            }
        

        // Allocate shared memory (actions) - Unix:
        #else        
            shm_id = shmget(DTMFBOX_SCRIPT_IN, sizeof(struct sh_actions_t), 0666);
            if(shm_id == -1) {
                app_err(THIS_FILE, "Cannot allocate shared memory!");
                return 1;
            }

            sh_actions_shmem = shmat(shm_id, 0, 0);
            if(!sh_actions_shmem)
            {
                app_err(THIS_FILE, "Cannot allocate memory for shared variable!");
                return 1;
            }
        #endif

    }

    return (sh_actions_shmem == NULL);
}

// Close or destroy shared memory variable (sh_actions_shmem)
int close_shared_memory(int destroy)
{
    #if DTMFBOX_WIN32
    {        
        if(destroy) {
            remove(DTMFBOX_SCRIPT_IN);
        }
    }
    #else
    {
        // destroy allocated shared memory
        if(sh_actions_shmem != NULL)
            shmdt(&sh_actions_shmem);

        sh_actions_shmem = NULL;

        // destroy the shared memory segment
        if(destroy)
        {
            struct shmid_ds shm_desc;

            if (shmctl(shm_id, IPC_RMID, &shm_desc) == -1) {
                app_err(THIS_FILE, "Cannot deallocate shared memory!");
                return 1;
            }
        }
    }
    #endif

    return 0;
}
