/*********************************************************************************
 * type_globals.h ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/

// --------------------------------------------------------------------------------------------
// global_t sub-types
// --------------------------------------------------------------------------------------------

// struct for registrar mode (connected clients)
typedef struct global_clients_t
{    
    char                contacts[MAX_CONTACTS][MAX_URI_LENGTH];   // contact(s) from client
    char                addr[MAX_URI_LENGTH];           // host:port from client        
    char                contact_param[MAX_URI_LENGTH];  // additional contact parameter    
    int                 contact_count;                  // contact count    
    int                 account_id[MAX_ACCOUNTS];       // account-id, the client registered with
    long                timeout;                        // Timeout before client will be removed from list
}
global_clients_t;

typedef struct global_conf_t
{
    pjmedia_conf        *bridge;                        // Conference bridge
    pjmedia_port        *bridge_port;                   // Conference port
    pjmedia_master_port *master_port;                   // Master port (slot 0)    
}
global_conf_t;

typedef struct global_audio_t
{
    pjmedia_snd_port    *snd_bidi_port;                 // Soundcard port for playing and recording (bidirectional)
    pjmedia_snd_port    *snd_play_port;                 // Soundcard port for playing
    pjmedia_snd_port    *snd_rec_port;                  // Soundcard port for recording
    pjmedia_port        *master_null_port;              // Null port for masterclock (when null_sound=1)
    pjmedia_port        *flow_null_port;                // Null port for connections (to keep the mediaflow running)    
    unsigned int        flow_null_slot;                 // Slot for the mediaflow

    char                snd_name_rec[64];               // Sounddevice name (microphone)
    char                snd_name_play[64];              // Sounddevice name (output)
}
global_audio_t;

// --------------------------------------------------------------------------------------------
// global_t main-type
// --------------------------------------------------------------------------------------------

typedef struct global_t
{
    pj_caching_pool     cp;                             // Global caching pool
    pj_pool_t           *pool;                          // Global memory pool    
    pj_mutex_t          *mutex;                         // Global mutex
    pj_mutex_t          *acc_mutex;                     // Account mutex (connection_by_free/reset_connection)
    pjsip_endpoint      *endpt;                         // Global SIP endpoint
    pjmedia_endpt       *med_endpt;                     // Global media endpoint
    
    pjsip_generic_string_hdr *useragent_hdr;            // User-Agent Header
    
    pj_sock_t           sock;                           // UDP server socket
    pj_stun_config      stun_cfg;                       // STUN config
    pj_sockaddr         stun_srv;                       // STUN server socket
    pj_status_t         stun_status;                    // STUN status
    pjsip_transport     *transport;                     // UDP server transport (PJSIP_TRANSPORT_UDP)
    pj_dns_resolver     *dns_resolver;                  // DNS resolver (SRV records)
    
    char                ip_addr_local[MAX_URI_LENGTH];  // Local IP
    char                ip_addr_global[MAX_URI_LENGTH]; // Global IP (STUN/ICE)
    int                 udp_port_local;                 // Local SIP UDP port
    int                 udp_port_global;                // Global SIP UDP port

    pj_thread_t         *voip_listen_thread;            // VoIP-listen loop thread
    pj_thread_t         *signal_handler_thread;         // signal_handler_thread for process communication        

    pj_timer_entry      keep_alive_timer;               // Keep-Alive timer entry
    pj_timer_entry      registrar_refresh_timer;        // Check clients for timeout (Registrar-Mode)
    pj_timer_entry      script_event_timer;             // Run script events

    char                current_exe[512];               // This binary (from startargs)
    char                workdir[512];                   // Working directory
    char                logfile[512];                   // Parameter: Logfile (-log)
    char                configfile[512];                // Parameter: Config file (-cfg)
    int                 closing;                        // When 1, it stops the listen-loop and exits application
    int                 daemonized;                     // Parameter: Daemon mode (-daemon)
    int                 foreground;                     // Parameter: Start in foreground (-start)
    int                 process_boosted;                // Priority boosted?

    int                 input_pipe;                     // A named pipe for commands 
    int                 output_pipe;            

    FILE*               file_log;                       // file pointer for logging
    
    global_conf_t       conf;                           // Conference bridge
    global_audio_t      audio;                          // Audio (nullsound/soundcard)
    global_clients_t    clients[MAX_CLIENTS];           // Authorized clients (Registrar-Mode)    
   
}
global_t;

#define DTMFBOX_LOCK()	    pj_mutex_lock(oGLB.mutex)
#define DTMFBOX_TRY_LOCK()  pj_mutex_trylock(oGLB.mutex)
#define DTMFBOX_UNLOCK()	pj_mutex_unlock(oGLB.mutex)

#define DTMFBOX_ACC_LOCK()	    pj_mutex_lock(oGLB.acc_mutex)
#define DTMFBOX_ACC_UNLOCK()	pj_mutex_unlock(oGLB.acc_mutex)
