/*********************************************************************************
 * type_sh_actions.h ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/

#define SCRIPT_MAX_QUEUE    1024                    // Max. number of simultaneously actions

// --------------------------------------------------------------------------------------------
// sh_actions_t sub-types
// --------------------------------------------------------------------------------------------

// internal queue
typedef struct sh_actions_queue_t
{
    int     ready;                                  // When 1, actions are going to be executed
    int     in_use;                                 // For interprocess-memory exchange (status: actions are filled, wait until not in use!)
    int     max_queue;                              // Max. queue no.
    int     cur_queue;                              // Current queue no.
    int     in_queue;                               // Count of instances in queue
    int     cur_pid;                                // Current process id
}
sh_actions_queue_t;

// return values
typedef struct sh_actions_return_t
{    
    char    environment[256];                       // Environment variable value (-env)
    int     play_id;                                // Playing ID (-play)    
}
sh_actions_return_t;

// -call
typedef struct sh_actions_call_t
{
    char    number[MAX_URI_LENGTH];                 // Number where call comes from
    char    target[MAX_URI_LENGTH];                 // Number or SIP-Uri to call
    int     capi_controller;                        // capi-controller to use (-call x x [ctrl])
    int     anonymous;                              // Anonymous call?
}
sh_actions_call_t;

// -play, -record
typedef struct sh_actions_play_rec_t
{
    char    play_file[256];                         // File/IP to play
    char    rec_file[256];                          // File/IP to record    
    int     play_thread[MAX_CONS][INCONF_PORTS];    // Play the wave/stream in foreground and wait until it ends (block).
    int     wave;                                   // Play/Record wave file?
    int     raw;                                    // Play/Record raw file (fifo/named pipe)?
    int     speex;                                  // Play/Record speex file?
    int     rtp;                                    // Send/Recv audio from/to RTP/RTCP stream?    
    int     rtp_port;                               // Port for RTP
    int     stream_clockrate;                       // Optional clockrate for raw file 
    long    stream_buffersize;                      // Optional buffersize for raw file
    int     wait_start;                             // Timeout (Stream-Start)
    int     wait_end;                               // Timeout (Stream-End)
   
}
sh_actions_play_rec_t;

// -mixer
typedef struct sh_actions_mixer_t
{
    int     volume;                                 // mixer volume
    int     fade_volume;                            // mixer fade volume
    int     fade_duration;                          // mixer fade duration (msec)
    int     options;                                // mixer options (volume, fade, background, talkover, etc..)
    int     play_id;                                // mixer control for play_id (-1 if not mixer option given)
    int     rec_id;                                 // mixer control for recording 
}
sh_actions_mixer_t;

// -timer
typedef struct sh_actions_timer_t
{
    char    optionstring[512];                      // timer commandline args
    long    timeout;                                // timer timeout (sec/msec - default: sec)    
    int     mode;                                   // 1=msec, 0=sec (default)?
}
sh_actions_timer_t;

// -bridge
typedef struct sh_actions_bridge_t
{
    int     connect;                                // Connect with other ID
    int     disconnect;                             // Disconnect other ID
}
sh_actions_bridge_t;

// -stop
typedef struct sh_actions_stop_t
{
    int     option;                                 // Stop play/record/daemon    
    int     play_id;                                // for stopping players by id (specified)            
}
sh_actions_stop_t;

// -scriptfile, -delimiter
typedef struct sh_actions_script_t
{
    char    file[256];                              // The scriptfile to change
    char    delimiter[2];                           // The delimiter  
    char    env_key[256];                           // Environment variable key name
    char    env_val[256];                           // Environment variable value
    int     waitevent;                              // Block script (wait for event)?
}
sh_actions_script_t;

// -tone
typedef struct sh_actions_tone_t
{
    short   active;                                 // Tone port active?
    short   freq1;                                  // Frequence 1
    short   freq2;                                  // Frequence 2
    short   on_msec;                                // On in msec
    short   off_msec;                               // Off in msec
    short   volume;                                 // Volume (0-32767)
}
sh_actions_tone_t;


// --------------------------------------------------------------------------------------------
// sh_actions_t main-type
// --------------------------------------------------------------------------------------------

typedef struct sh_actions_t
{   
    int                        connection;          // Connection-ID    
        
    sh_actions_call_t          call;                // -call
    sh_actions_script_t        script;              // -scriptfile, -delimter
    sh_actions_play_rec_t      play_rec;            // -play, -record
    sh_actions_timer_t         timer;               // -timer
    sh_actions_stop_t          stop;                // -stop       
    sh_actions_tone_t          tone;                // -tone
    sh_actions_mixer_t         mixer;               // -mixer
    sh_actions_bridge_t        bridge;              // -bridge    
    int                        hook;                // -hook
    char                       dtmf[16];            // -dtmf
    char                       text[33];            // -text
    int                        list;                // -list (connections, clients, accounts)    

    int                        valid_commands;      // Has valid commands?
    int                        parse_plugin;        // Parse command line to the loaded plugins
    char                       args[256][256];
    int                        argc;

    sh_actions_queue_t         queue;               // internal queue 
    sh_actions_return_t        retval;              // return values
    
    connection_info_t          con_info[MAX_CONS];  // connection info (gets updated by thread)
}
sh_actions_t;
