/*********************************************************************************
 * voip_events.c ($Revision: 189 $)
 *
 * Copyright (c) 2006-2008 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "voip_events.c"

#if DTMFBOX_USE_VOIP

// DTMF callback from the stream
void dtmf_callback(pjmedia_stream *strm, void *user_data, int digit)
{
    unsigned id = (int)user_data;
    PJ_UNUSED_ARG(strm);

    voip_get_dtmf((char)digit, 0, id);
}

//
// Callback when INVITE session state has changed.
// This callback is registered when the invite session module is initialized.
// Here we get the information about the session state (disconnected, connecting, confirmed, early)
//
void call_on_state_changed( pjsip_inv_session *inv, pjsip_event *e)
{
    int i, j;

    PJ_UNUSED_ARG(e);

    // Get the invite session from connections
    i = connection_by_invite(inv);

    switch(inv->state)
    {       
        // ----------------------------------------------------------
        // State: CONNECTING
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_CONNECTING:
            app_msg(THIS_FILE, "Call CONNECTING [port=%d, reason=%d (%s)]", i, inv->cause, pjsip_get_status_text(inv->cause)->ptr);
            break;

        // ----------------------------------------------------------
        // State: EARLY
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_EARLY:
            app_msg(THIS_FILE, "Call EARLY [port=%d, reason=%d (%s)]", i, inv->cause, pjsip_get_status_text(inv->cause)->ptr);
            
            // Script event 
            if(i != -1) {                
                // no client (CONNECT), incoming only
                if(!(oCONS[i].info->direction & CLIENT)) {
                    script_event_add(VOIP, CONNECT, oCONS[i].script->file, oCONS[i].id, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");
                }

                // client (DDI/EARLY), outgoing only
                else if((oCONS[i].info->direction & CLIENT) && !(oCONS[i].info->direction & INCOMING)) 
                {
                    unsigned j;
                    char to[MAX_URI_LENGTH];

                    // CONNECT
                    script_event_add(VOIP, CONNECT, oCONS[i].script->file, i, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");

                    // DDI
                    strncpy(to, oCONS[i].info->no_target, sizeof(to));
                    sip_strip_user(to);
                    for(j=0; j<strlen(to); j++)
                        voip_get_dtmf(to[j], 1, i);                

                    // EARLY
                    if(!oCONS[i].capi->SetupEarly) {
                        script_event_add(VOIP, EARLY, oCONS[i].script->file, oCONS[i].id, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");
                        oCONS[i].capi->SetupEarly = 1;
                    }
                                                         
                    //pj_thread_sleep(2000);
                }
            }
            break;

        // ----------------------------------------------------------
        // State: CONFIRMED
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_CONFIRMED:
            app_msg(THIS_FILE, "Call CONFIRMED [port=%d, reason=%d (%s)]", i, inv->cause, pjsip_get_status_text(inv->cause)->ptr);
            if(i != -1)
            {
                // Status: CONFIRMED
                oCONS[i].info->status = 2;

                // Initialize new tone port
                if(oCONS[i].ports->tone == NULL)            
                    port_tone_init(&oCONS[i], 0, 0, 0, 0, 0);
                
                    
                for(j=0; j<INCONF_PORTS; j++)
                {
                    int in_conf = oCONS[i].in_conf[j];

                    // Registrar-Mode: hook up client?
                    if(in_conf != -1) 
                    {         
                        if(oCONS[i].info->direction & CLIENT || (oCONS[in_conf].info->direction & CLIENT && oCONS[i].info->active))
                        {                                             
                            if(oCONS[in_conf].info->type == VOIP && !(oCONS[in_conf].info->direction & OUTGOING) ||
                               oCONS[in_conf].info->type == VOIP && oCONS[i].info->status == 2 && !(oCONS[i].info->direction & CLIENT)) 
                            {
                                voip_confirm_call(&oCONS[in_conf]);
                            }

                            if(oCONS[in_conf].info->type == CAPI) {
                                capi_confirm_call(&oCONS[in_conf]);
                            }

                            // disconnect all unactive clients
                            if(!(oCONS[in_conf].info->direction & CLIENT)) 
                                voip_registrar_disconnect_clients(&oCONS[in_conf], 1);
                            if(!(oCONS[i].info->direction & CLIENT)) 
                                voip_registrar_disconnect_clients(&oCONS[i], 1);
                        }
                    }
                

               }
                
                // SCRIPT: "CONFIRMED"
                script_event_add(VOIP, CONFIRMED, oCONS[i].script->file, i, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");
            }
            break;

        // ----------------------------------------------------------
        // State: DISCONNECTED
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_DISCONNECTED:
            app_msg(THIS_FILE, "Call DISCONNECTED [port=%d, reason=%d (%s)]", i, inv->cause, pjsip_get_status_text(inv->cause)->ptr);

            if(i!=-1)
            {        
                // Registrar-Mode: disconenct every client
                voip_registrar_disconnect_clients(&oCONS[i], 0);

                // hangup current call                                
                if(!voip_hangup(&oCONS[i]))
                    app_dbg(THIS_FILE, "Hangup [%d]", i);
                
                // Reset transport
                oCONS[i].rtp->g_inv = NULL;
                oCONS[i].rtp->g_med_session = NULL;
                oCONS[i].rtp = NULL;

                // Reset current connection
                reset_connection(&oCONS[i]);                                    
            }

            break;

        // ----------------------------------------------------------
        // State: unknown
        // ----------------------------------------------------------
        default:
            app_msg(THIS_FILE, "Call state changed to %s [port=%d]", pjsip_inv_state_name(inv->state), i);
            break;
    }
}


//
// Callback when SDP negotiation has completed.
// This callback initialize the media as soon the negotation has completed.
//
void call_on_media_update(pjsip_inv_session *inv, pj_status_t status)
{
    pjmedia_session_info sess_info;
    const pjmedia_sdp_session *local_sdp;
    const pjmedia_sdp_session *remote_sdp;
    unsigned int i, j;

    app_dbg(THIS_FILE, "On media update");

    if (status != PJ_SUCCESS)
    {
        app_err(THIS_FILE, "(%d) SDP negotiation has failed", status);
        return;
    }

    // Get invite session from connections
    i = connection_by_invite(inv);
    if(i == -1)
    {
        app_err(THIS_FILE, "Can not find invite session");
        return;
    }

    // mhh...
    if(oCONS[i].info->active == 1)
    {
        // app_err(THIS_FILE, "(%d) already connected", status);
        // return;
    }

    // Get local and remote SDP.
    status = pjmedia_sdp_neg_get_active_local(inv->neg, &local_sdp);
    status = pjmedia_sdp_neg_get_active_remote(inv->neg, &remote_sdp);

    // Create session info
    status = pjmedia_session_info_from_sdp(inv->pool, oGLB.med_endpt, 1, &sess_info, local_sdp, remote_sdp);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create media session", status);
        return;
    }


#if DTMFBOX_USE_ICE
#if (PJSIP_VERSION == 8)
    // ICE: Start ICE media flow
    if(oSET.voip.ice)
    {
        status = pjmedia_ice_start_ice(oCONS[i].rtp->g_med_transport, inv->pool, remote_sdp, 0);
        if(status != PJ_SUCCESS) {
            app_err(THIS_FILE, "(%d) Unable to start ICE media session", status);
            return;
        }
    }
#else     
     status = pjmedia_transport_media_start(oCONS[i].rtp->g_med_transport, inv->pool, local_sdp, remote_sdp, 0); 
     if(status != PJ_SUCCESS) {
            app_err(THIS_FILE, "(%d) Unable to start media transport", status);
     }
#endif
#endif


    // Stream settings
    for(j=0; j<sess_info.stream_cnt; j++)
    {
        char info[512];
        int len;

        // Display current codec
        pj_memset(info, 0, sizeof(info));
        len = pj_ansi_snprintf(info, sizeof(info), "stream #%d: %.*s (%d/%d)", j, (int)sess_info.stream_info[j].fmt.encoding_name.slen, sess_info.stream_info[j].fmt.encoding_name.ptr, sess_info.stream_info[j].fmt.clock_rate, sess_info.stream_info[j].fmt.channel_cnt);
        app_dbg(THIS_FILE, "%s", info);

        // Disable VAD for each stream (silence detection)
        if(!oSET.voip.use_vad)
        {
            app_msg(THIS_FILE, "stream #%d: disable VAD [port=%d]", j, i);
            sess_info.stream_info[j].param->setting.vad = 0;
        }
/*
// PTIME
        sess_info.stream_info[j].param->setting.frm_per_pkt = (pj_uint8_t)
		1;
	    if (sess_info.stream_info[j].param->setting.frm_per_pkt == 0)
		sess_info.stream_info[j].param->setting.frm_per_pkt = 1;
*/
        // >> CHANGE OTHER CODEC AND STREAM SETTINGS HERE <<
    }


    // Create new media session. It is active immediately.
    status = pjmedia_session_create(oGLB.med_endpt, &sess_info, &oCONS[i].rtp->g_med_transport, NULL, &oCONS[i].rtp->g_med_session);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create media session", status);
        return;
    }
    
    // Get the media port interface from the first stream in the session, so we
    // have a basicly struct containing get_frame/put_frame.
    // This media port can be attached to the conference bridge or to a sound player/recorder device.
    // It is also needed by capi_streaming to stream capi<->voip.
    status = pjmedia_session_get_port(oCONS[i].rtp->g_med_session, 0, &oCONS[i].ports->media);
    
    if(status == PJ_SUCCESS)
    {    
        // Connection is ACTIVE and CONFIRMED
        oCONS[i].info->active = 1;
        oCONS[i].info->status = 2;
        
        // Wait for ringing thread 	
        if(oCONS[i].tone_thread != NULL && oCONS[i].ports->tone != NULL) {
            pj_thread_join(oCONS[i].tone_thread);
            pj_thread_destroy(oCONS[i].tone_thread);
            oCONS[i].tone_thread = NULL;
        }
                        
        // Initialize custom voip_port (second time) (for putframe/getframe)            
        // Now we have a media_port!
        port_voip_init(&oCONS[i]);

        // Connect conference partner, if there is one (media-streaming: voip<->voip)
        for(j=0; j<INCONF_PORTS; j++)
        {
          if(oCONS[i].in_conf[j] != -1)
          {
              mixer_conf_port_connect(&oCONS[i], &oCONS[oCONS[i].in_conf[j]], 1);
              mixer_conf_port_connect(&oCONS[oCONS[i].in_conf[j]], &oCONS[i], 1);
          }
        }
    }
    else {
       app_err(THIS_FILE, "(%d) Unable to get media port!", status);
       return;
    }

    // Set DTMF callback
    status = pjmedia_session_set_dtmf_callback(oCONS[i].rtp->g_med_session, 0, &dtmf_callback, (void*)i);
    if(status != PJ_SUCCESS)
        app_err(THIS_FILE, "(%d) Unable to initialize dtmf callback!", status);

}

// This callback is called when dialog has forked.
void call_on_forked(pjsip_inv_session *inv, pjsip_event *e)
{
    PJ_UNUSED_ARG(inv);
    PJ_UNUSED_ARG(e);
}

/*
// This callback is called, when tsx session-state has changed.
// At the moment, we use it for a workaround to receive DTMF by PJSIP_OTHER_METHOD
void call_on_tsx_changed(pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_event *e)
{    
    pjsip_method *method;
    int con_id = 0;
    pj_status_t status;

    PJ_UNUSED_ARG(inv);

    method = &tsx->method;
    con_id = connection_by_invite(inv);

    // DTMF workaround (some providers)
    //
    // we receive OTHER method with status_code=0 on DTMF so PRI disconnects :(
    // we send 200 response and call the dtmf-callback manually...
    if(method->id == PJSIP_OTHER_METHOD && tsx->status_code == 0)
    {
        char info_msg[128];
        pjsip_dialog *dlg;
        pjsip_tx_data *tdata;
        const pj_str_t reason = { "DTMF", 4 };

        // Get dialog
        dlg = pjsip_tsx_get_dlg(tsx);

        // Send 200 Ok response (so we do not get disconnected)
        if(e->body.rx_msg.rdata->msg_info.msg->type == PJSIP_REQUEST_MSG)
        {
            tsx->status_code = 200;
            status = pjsip_dlg_create_response(dlg, e->body.rx_msg.rdata, 200, &reason, &tdata);
            if (status == PJ_SUCCESS) {
                status = pjsip_dlg_send_response(dlg, tsx, tdata);
            }

            // Fetch body and read signal
            // pass dtmf to callback-function
            pjsip_print_text_body(e->body.rx_msg.rdata->msg_info.msg->body, (char*)&info_msg, sizeof(info_msg));
            if(strstr(info_msg, "Signal=") && strlen(info_msg) >= 8)
            {
                char dtmf[2];
                dtmf[0] = info_msg[7];
                dtmf_callback(NULL, (void*)con_id, (int)dtmf[0]);
            }
        }
    }    
}
*/

void registrar_connect_outgoing(void** args)
{
    char data[MAX_URI_LENGTH*5];
    char SourceNo[MAX_URI_LENGTH];
    char TargetNo[MAX_URI_LENGTH];    
    int con_id;
    int acc_id;
    int client_id;

    memset(data, 0, sizeof(data));
    memcpy(data, args, sizeof(data));
        
    con_id = ((int)(data[MAX_URI_LENGTH*0]));
    snprintf(SourceNo, MAX_URI_LENGTH, "%s", &data[MAX_URI_LENGTH*1]);
    snprintf(TargetNo, MAX_URI_LENGTH, "%s", &data[MAX_URI_LENGTH*2]);    
    acc_id = ((int)(data[MAX_URI_LENGTH*3]));    
    client_id = ((int)(data[MAX_URI_LENGTH*4]));
    
    pj_thread_sleep(2000);
       
    // MAKE CALL
    if(!oCONS[con_id].info->active && oCONS[con_id].info->status == 1) 
    {
        voip_registrar_connect_clients(&oCONS[con_id], SourceNo, TargetNo, acc_id, client_id, oSET.accounts[acc_id].controller.device, OUTGOING);
    }
}

//
// Callback when incoming requests outside any transactions and any
// dialogs are received.
// We're only interested to hande incoming INVITE, REGISTER and OTHER
// request, and we'll reject any other requests with 500 response.
//
pj_bool_t on_rx_request(pjsip_rx_data *rdata)
{    
    pjsip_dialog *dlg;
    pjmedia_sdp_session *local_sdp;
    pjsip_msg *msg = rdata->msg_info.msg;
    pjsip_tx_data *tdata;    
    pj_status_t status;
    int i,j,x;
    unsigned int options = 0;
    pjsip_fromto_hdr *hdr_from;
    pjsip_fromto_hdr *hdr_to;
    pjsip_sip_uri *uri_from;
    pjsip_sip_uri *uri_to;
    char from_hdr[MAX_URI_LENGTH];
    char to_hdr[MAX_URI_LENGTH];    
    int VOIP_CONNECTION1 = -1;
    int VOIP_CONNECTION2 = -1;
    int VOIP_RTP = -1;

    int found = 0;
    pjsip_uri *uri;
    pjsip_sip_uri *sip_uri;
    char from[MAX_URI_LENGTH];
    char to[MAX_URI_LENGTH];
    char host[MAX_URI_LENGTH];    
    char tmp[MAX_URI_LENGTH];
    pj_str_t reason;

    char SourceNo[MAX_URI_LENGTH]; 
    char TargetNo[MAX_URI_LENGTH];

    int incoming_direct = 0;
    int incoming_account = 0;    
    int incoming_registrar = 0;
    int outgoing_registrar = 0;
    int account_id = -1;
    int client_id = -1;

    // Status
    app_dbg(THIS_FILE, "RX-REQUEST %d bytes %s from %s %s:%d", rdata->msg_info.len, pjsip_rx_data_get_info(rdata), rdata->tp_info.transport->type_name, rdata->pkt_info.src_name, rdata->pkt_info.src_port);
   
    // Show RX-Message
    app_dbg2(THIS_FILE, "RX-SIP:\n\n%.*s\n", rdata->msg_info.len, rdata->msg_info.msg_buf);

    // REGISTRAR?
    if(voip_registrar_authorization_challenge(rdata))
        goto hell;
        //return PJ_FALSE;

    // PUBLISH?
    if(rdata->msg_info.msg->line.req.method.id == PJSIP_OTHER_METHOD && oSET.voip.registrar)
    {            
        pj_str_t reason = pj_str("Method not allowed!");
        pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 405, &reason, NULL, NULL);
        app_err(THIS_FILE, "Method not allowed (%d)!", rdata->msg_info.msg->line.req.method.id);
        goto hell;
        //return PJ_FALSE;
    }
    
    // Only INVITE-Requests. Any other requests will be rejected...
    if(rdata->msg_info.msg->line.req.method.id != PJSIP_INVITE_METHOD)
    {
        //pj_str_t reason = pj_str("Request not supported!");
        //pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 500, &reason, NULL, NULL);
        app_err(THIS_FILE, "Request not supported (%d)!", rdata->msg_info.msg->line.req.method.id);
        goto hell;
        //return PJ_FALSE;
    }

    // Got dialog or tsx?
    if (pjsip_rdata_get_dlg(rdata) || pjsip_rdata_get_tsx(rdata))
        goto hell;
	    //return PJ_FALSE;

    // Verify that we can handle the request  
    options = PJSIP_INV_SUPPORT_100REL;
    status = pjsip_inv_verify_request(rdata, &options, NULL, NULL, oGLB.endpt, NULL);
    if (status != PJ_SUCCESS)
    {
        app_err(THIS_FILE, "Sorry, can not handle this invite!");        
        goto hell;
        //return PJ_FALSE;
    }

    // --------------------------------------------------------------------------------------------------------
    // Check, if this is an incoming or outgoing connection.
    // The FROM-Header contains the address of the client.
    // We search for the address in the list of all registred clients (oGLB.clients).
    // When the client was found, the connection is an outgoing connection (through this regstrar).
    // --------------------------------------------------------------------------------------------------------        
    // Search for client
    for(i=0; i<oSET.voip.max_clients; i++)
    {        
        pj_ansi_snprintf(tmp, sizeof(tmp), "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port);
        //if(!pj_strcmp2(&oGLB.clients[i].addr, (const char*)&tmp))
        if(!strcmp(oGLB.clients[i].addr, tmp))
        {            
            // Outgoing request (Registrar-Mode only!)
            // Address was found in the clients list!
            client_id = i;            
            found = 1;  
            break;
        }
    }
    if(!found)
    {
        // An incoming request (Client-Mode and Registrar-Mode!)
    }

    // --------------------------------------------------------------------------------------------------------
    // The source and destination address of the sip message must be changed:
    //
    // On outgoing connections, we use the "real" registrar as destination address.
    // On incoming connections, we use the client's address as destination. The phone rings...
    // --------------------------------------------------------------------------------------------------------
    hdr_from = (pjsip_fromto_hdr*) pjsip_msg_find_hdr(msg, PJSIP_H_FROM, NULL);
    hdr_to = (pjsip_fromto_hdr*) pjsip_msg_find_hdr(msg, PJSIP_H_TO, NULL);
    uri_from = (pjsip_sip_uri*) pjsip_uri_get_uri(hdr_from->uri);
    uri_to = (pjsip_sip_uri*) pjsip_uri_get_uri(hdr_to->uri);
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_from, from_hdr, MAX_URI_LENGTH);
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_to, to_hdr, MAX_URI_LENGTH);

    sip_strip_user((char*)&from_hdr);
    sip_strip_user((char*)&to_hdr);

    // INCOMING
    if(!found)
    {
        // Search for account
        account_id = voip_get_acc_incoming(rdata, from, to);            
        
        // When this account is used as registrar account, search for any registered clients!
        if(account_id != -1)
        {
            int client_acc_id = -1;

            for(i=0; i<MAX_ACCOUNTS; i++)
            {                    
               client_acc_id = client_account_by_account(i, account_id);
               if(client_acc_id != -1) 
                   break;                   
            }

            if(client_acc_id != -1) 
            {
                // a client found! Set as incoming registrar                
                incoming_registrar = 1;                    
            }
            else 
            {
                // no client found! Just set as incoming client                
                incoming_account = 1;
            }
        }
        else
        {            
            // direct IP call (no account found)
            incoming_direct = 1;  
            
            // unknown accounts allowed?
            if(!oSET.voip.allow_unknown_acc)
            {
                reason = pj_str("Unknown number!");
                pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 400, &reason, NULL, NULL);
                app_msg(THIS_FILE, "Incoming call for unknown account [%s -> %s] (ignoring)", from, to);
                goto hell;
                //return PJ_FALSE;
            }
        }

        app_msg(THIS_FILE, "Incoming call from %s to %s [account #%d]", from, to, account_id);
    }

    // OUTGOING
    else
    {
        account_id = voip_get_acc_outgoing(rdata, from, to);
        if(account_id == -1) 
        {
            app_err(THIS_FILE, "UNAUTHORIZED OUTGOING CLIENT REQUEST!");
            goto hell;
        }        

        outgoing_registrar=1;
    }
  
    // --------------------------------------------------------------------------------------------------------
    // Get a free connection for this request.
    // --------------------------------------------------------------------------------------------------------    
    DTMFBOX_ACC_LOCK();
    VOIP_CONNECTION1 = connection_by_free();                                                // 1. Connection
    if(VOIP_CONNECTION1 == -1) {
        DTMFBOX_ACC_UNLOCK();
        goto hell;
    }
        
    // An INVITE-Session already exist? Abort!
    if (oCONS[VOIP_CONNECTION1].rtp)
    {
        app_err(THIS_FILE, "Another call is in progress!");
        reason = pj_str("Another call is in progress!");
        pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 500, &reason, NULL, NULL);        
        goto hell;
    } 
    else 
    {
        VOIP_RTP = transport_by_free();
        if(VOIP_RTP == -1)
        {
            app_err(THIS_FILE, "No free transports left. Max. %d!", oSET.voip.rtp_max);
            reason = pj_str("No free transports left!");
            pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 500, &reason, NULL, NULL);        
            DTMFBOX_ACC_UNLOCK();
            goto hell;
        }        
        
        oCONS[VOIP_CONNECTION1].rtp = oRTP[VOIP_RTP];
        oCONS[VOIP_CONNECTION1].info->type = VOIP;
    }
    DTMFBOX_ACC_UNLOCK();
    
    // 1. Connection: fill connection details ...        
    oCONS[VOIP_CONNECTION1].capi->NCCI = (unsigned)-1;                                   // 1. Connection: Pseudo NCCI (-1 = VoIP connection)
    oCONS[VOIP_CONNECTION1].info->account_id = account_id;                               // 1. Connection: Account-ID    
    oCONS[VOIP_CONNECTION1].info->status = 0;
    
    // Some user agents urlencode characters.
    // make sure, that # is decoded!!
    replace(from, from, sizeof(from), "%23", "#");
    replace(to, to, sizeof(to), "%23", "#");

    if(!found) {
        oCONS[VOIP_CONNECTION1].info->direction = INCOMING;
        pj_ansi_snprintf(oCONS[VOIP_CONNECTION1].info->no_source, sizeof(oCONS[VOIP_CONNECTION1].info->no_source), "%s", to);
        pj_ansi_snprintf(oCONS[VOIP_CONNECTION1].info->no_target, sizeof(oCONS[VOIP_CONNECTION1].info->no_target), "%s", from);

    } else {
        oCONS[VOIP_CONNECTION1].info->direction = CLIENT | OUTGOING;    
        pj_ansi_snprintf(oCONS[VOIP_CONNECTION1].info->no_source, sizeof(oCONS[VOIP_CONNECTION1].info->no_source), "%s", from);
        pj_ansi_snprintf(oCONS[VOIP_CONNECTION1].info->no_target, sizeof(oCONS[VOIP_CONNECTION1].info->no_target), "%s", to);
    }

    pj_ansi_snprintf(oCONS[VOIP_CONNECTION1].script->file, sizeof(oCONS[VOIP_CONNECTION1].script->file), "%s", oSET.scriptfile);
    pj_ansi_snprintf(oCONS[VOIP_CONNECTION1].script->delimiter, sizeof(oCONS[VOIP_CONNECTION1].script->delimiter), "%s", oSET.delimiter);
    
    // Create UAS dialog
    app_dbg2(THIS_FILE, "Create UAS dialog...");    

    {                
        pj_str_t contact;

        // get suitable contact
        status = voip_create_uas_contact(oCONS[VOIP_CONNECTION1].pools->con_pool, &contact, account_id, rdata); 
        if(status == PJ_SUCCESS)
            status = pjsip_dlg_create_uas(pjsip_ua_instance(), rdata, &contact, &dlg);
        else            
            status = pjsip_dlg_create_uas(pjsip_ua_instance(), rdata, NULL, &dlg);
        
        if (status != PJ_SUCCESS) 
        {
            app_err(THIS_FILE, "Unable to create UAS dialog!");
            pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 500, NULL, NULL, NULL);
            goto hell;
        }
    }    

    #if DTMFBOX_USE_ICE
    #if (PJSIP_VERSION == 8)
        // ICE: restart
        if(oSET.voip.ice)
        {
            pj_ice_sess_role ice_role = PJ_ICE_SESS_ROLE_CONTROLLED; // UAS            
                                   
		    // Get media socket info for transport!
            pjmedia_transport_get_info(oCONS[VOIP_CONNECTION1].rtp->g_med_transport, &oCONS[VOIP_CONNECTION1].rtp->g_med_skinfo);
					   
            // Restart delicious ICE ;)
            pjmedia_ice_stop_ice(oCONS[VOIP_CONNECTION1].rtp->g_med_transport);
            status = pjmedia_ice_init_ice(oCONS[VOIP_CONNECTION1].rtp->g_med_transport, ice_role, NULL, NULL);
        }
    #endif
    #endif

    // Create SDP capability and get media socket info!    
    #if PJSIP_VERSION == 8
        pjmedia_transport_get_info(oCONS[VOIP_CONNECTION1].rtp->g_med_transport, &oCONS[VOIP_CONNECTION1].rtp->g_med_skinfo);
    #else
        {
            pjmedia_transport_info tpinfo;

            pjmedia_transport_get_info(oCONS[VOIP_CONNECTION1].rtp->g_med_transport, &tpinfo);
            pj_memcpy(&oCONS[VOIP_CONNECTION1].rtp->g_med_skinfo, &tpinfo.sock_info, sizeof(pjmedia_sock_info));
        }
    #endif
    
    app_dbg2(THIS_FILE, "Create media capability...");
    status = pjmedia_endpt_create_sdp(oGLB.med_endpt, rdata->tp_info.pool, 1, &oCONS[VOIP_CONNECTION1].rtp->g_med_skinfo, &local_sdp);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "Can not create media capability!");
        goto hell;
    }

    #if DTMFBOX_USE_ICE
    #if (PJSIP_VERSION == 8)
        // ICE: modify SDP
        if(oSET.voip.ice)
        {            
            status = pjmedia_ice_modify_sdp(oCONS[VOIP_CONNECTION1].rtp->g_med_transport, oCONS[VOIP_CONNECTION1].pools->con_pool, local_sdp);
            if(status != PJ_SUCCESS)
            {
                app_err(THIS_FILE, "Can not modify SDP for ICE!");
                goto hell;
            }
        }
    #else
        status = pjmedia_transport_media_create(oCONS[VOIP_CONNECTION1].rtp->g_med_transport, oCONS[VOIP_CONNECTION1].pools->con_pool, 0, local_sdp, 0);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create media transport!", status);
            goto hell;
        }
    #endif
    #endif


    // Create invite session
    app_dbg2(THIS_FILE, "Create invite session..."); 
    status = pjsip_inv_create_uas(dlg, rdata, local_sdp, PJSIP_INV_SUPPORT_100REL, &oCONS[VOIP_CONNECTION1].rtp->g_inv);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create invite session!", status);
        goto hell;
    }

    // Create dialog 100 response
    app_dbg2(THIS_FILE, "Respond 100...");
    status = pjsip_dlg_create_response(dlg, rdata, 100, NULL, &tdata);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create 100 respond!", status);
        goto hell;
    }
    
    // Send initial 100 response
    status = pjsip_dlg_send_response(dlg, pjsip_rdata_get_tsx(rdata), tdata);
    if(status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to send 100 respond!", status);
        goto hell;
    }

    // --------------------------------------------------------------------------------------------------------
    // First respond with code 180 to initialize invite (ring)
    // (pjsip_inv_initial_answer !!!)
    // --------------------------------------------------------------------------------------------------------
    app_dbg2(THIS_FILE, "Respond 180 (*ring*)..");
    status = pjsip_inv_initial_answer(oCONS[VOIP_CONNECTION1].rtp->g_inv, rdata, 180, NULL, NULL, &tdata);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create 180 respond!", status);
        goto hell;
    }

    // Send the 180 response.    
    status = pjsip_inv_send_msg(oCONS[VOIP_CONNECTION1].rtp->g_inv, tdata);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to send 180 respond!", status);
        goto hell;
    }

    // --------------------------------------------------------------------------------------------------------
    // INCOMING (Client-Mode):
    // --------------------------------------------------------------------------------------------------------
    if((incoming_account || incoming_direct) && (!incoming_registrar && !outgoing_registrar)) {
        return PJ_TRUE;
    }

    // --------------------------------------------------------------------------------------------------------
    // INCOMING/OUTGOING (Registrar-Mode):
    //
    // On incoming request, we call each registered client.
    // On outgoing request, we just make a call and wait for connect 
    // --------------------------------------------------------------------------------------------------------
    if(oSET.voip.registrar)
    {   
        // REGISTRAR: INCOMING CALL
        //if(!found)
        if(incoming_registrar)
        {               
            x = 0;

            // Search for clients
            for(i=0; i<oSET.voip.max_clients; i++)
            {
                // Search for account
                for(j=0; j<MAX_ACCOUNTS; j++)
                {                    
                    int acc_id = oGLB.clients[i].account_id[j];

                    // Account and client found? Call!                    
                    if(acc_id == account_id)
                    {
                        // Get sip-uri
                        uri = rdata->msg_info.to->uri;
                        sip_uri = (pjsip_sip_uri*)pjsip_uri_get_uri(uri);

                        if (PJSIP_URI_SCHEME_IS_SIP(uri))
                        {
                            char ip[MAX_URI_LENGTH];
                            int port;
                            
                            pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, sip_uri, to, MAX_URI_LENGTH);
                            pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, sip_uri, host, MAX_URI_LENGTH);
                            sip_strip_user((char*)&to);
                            sip_strip_host((char*)&host);
                            
                            pj_ansi_snprintf(ip, sizeof(ip), "%s", oGLB.clients[i].addr);                            
                            pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "sip:%s@%s", oSET.accounts[acc_id].registrar_username, oGLB.clients[i].addr);
                                                        
                            // public or private ip?
                            // change host SourceNo, so that media flow is correctly (private/public ip)
                            {                                
                                strcpy(to, oCONS[VOIP_CONNECTION1].info->no_target);
                                sip_strip_user(to);
                                /*                                
                                if(!ip_is_public(ip)) {
                                    pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s@%s:%d", to, oGLB.ip_addr_local, oGLB.udp_port_local);
                                } else {
                                */
                                    //pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s@%s:%d", to, oGLB.ip_addr_global, oGLB.udp_port_global);
                                
                                    status = voip_resolve_transport_address(ip, &port);
                                    if(status != PJ_SUCCESS)
                                    {
                                        app_err(THIS_FILE, "(%d) Unable to resolve transport address", status);
                                        break;
                                    }                                    
                                    pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s@%s:%d", to, ip, port);
                                //}

                                VOIP_CONNECTION2 = voip_registrar_connect_clients(&oCONS[VOIP_CONNECTION1], SourceNo, TargetNo, account_id, i, VOIP, INCOMING | CLIENT);
                            }
                            break;
                        }
                    }
                }
            }
        }

        // REGISTRAR: OUTGOING CALL
        else if(outgoing_registrar)
        {               
            // INIT OUTGOING VOIP CALL -->
            if(oSET.accounts[account_id].controller.device == VOIP)
            {
                pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s", oSET.accounts[account_id].id);
                pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "sip:%s", to);                
            }
            
            // INIT OUTGOING CAPI CALL -->
            else if(oSET.accounts[account_id].controller.device == CAPI)
            {                
                sip_strip_user(to);
                pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "%s", oSET.accounts[account_id].number);                
                pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "%s", to);                         
            }
                     
            {
                char args[MAX_URI_LENGTH*5];
                memset(args, 0, sizeof(args));
                pj_memcpy(&args[MAX_URI_LENGTH*0], &VOIP_CONNECTION1, MAX_URI_LENGTH);
                pj_memcpy(&args[MAX_URI_LENGTH*1], SourceNo, MAX_URI_LENGTH);
                pj_memcpy(&args[MAX_URI_LENGTH*2], TargetNo, MAX_URI_LENGTH);
                pj_memcpy(&args[MAX_URI_LENGTH*3], &account_id, MAX_URI_LENGTH);
                pj_memcpy(&args[MAX_URI_LENGTH*4], &client_id, MAX_URI_LENGTH);
                pj_thread_create(oCONS[VOIP_CONNECTION1].pools->con_pool, "registrar_connect_%p", (pj_thread_proc*)&registrar_connect_outgoing, (void*)&args, 0, (unsigned)NULL, &oCONS[VOIP_CONNECTION1].tone_thread);
                pj_thread_sleep(250);
            }

            /*
            // MAKE CALL
            if(!oCONS[VOIP_CONNECTION1].info->active && oCONS[VOIP_CONNECTION1].info->status == 1) 
            {
                VOIP_CONNECTION2 = voip_registrar_connect_clients(&oCONS[VOIP_CONNECTION1], SourceNo, TargetNo, account_id, client_id, oSET.accounts[account_id].controller.device, OUTGOING);
            }
            */
        }
    }
    else
    {
        app_err(THIS_FILE, "Registrar- and client-mode inactive! You must activate one or both mode(s)!");
        goto hell;
    }




// ok!
// ---------------------------------------------------------------------------------------
return PJ_TRUE;


// Error! Reset connection(s)...
// ---------------------------------------------------------------------------------------
hell:
    // Reset connections
    if(VOIP_CONNECTION1 >= 0)
        reset_connection(&oCONS[VOIP_CONNECTION1]);

    if(VOIP_CONNECTION2 >= 0)
        reset_connection(&oCONS[VOIP_CONNECTION2]);

    return PJ_TRUE;
}

#endif

